# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import numpy as np
import os
import shutil
from qonnx.custom_op.general import im2col
from qonnx.custom_op.general.im2col import compute_conv_output_dim
from qonnx.custom_op.registry import getCustomOp
from qonnx.util.basic import roundup_to_integer_multiple

from finn.custom_op.fpgadataflow.convolutioninputgenerator import ConvolutionInputGenerator
from finn.custom_op.fpgadataflow.rtlbackend import RTLBackend

# RTL Convolution Input Generator / Sliding Window Generator (SWG)
# Matches and extends the functionality of all ConvolutionInputGenerator_* functions
# in finn-hlslib by generating HDL code for two different implementation styles:
# - Addressable cyclic buffer: to be used when out_width <= in_width
# - Parallel registers + line buffers: to be used when out_width > in_width
# Supports non-square, 1D, strided, dilated, and depthwise convolutions.
# Note: the actual data layout produced is different for depthwise and non-depthwise:
# * non-depthwise SWG: (1, OFMDim_H, OFMDim_W, K_H, K_W, IFMChannels/SIMD, SIMD)
# * depthwise SWG: (1, OFMDim_H, OFMDim_W, IFMChannels/SIMD, K_H, K_W, SIMD)

# NOTE: "Parallel" implementation style not yet implemented in this version!


class ConvolutionInputGenerator_rtl(ConvolutionInputGenerator, RTLBackend):
    """Class that corresponds to finn-rtllib swg module.
    Generates an RTL ConvolutionInputGenerator implementation
    based on (System-)Verilog templates, defined in finn-rtllib/swg."""

    def __init__(self, onnx_node, **kwargs):
        super().__init__(onnx_node, **kwargs)

    def get_nodeattr_types(self):
        my_attrs = {
            # additional parallelization parameter - not yet implemented
            "M": ("i", False, 1),
        }
        my_attrs.update(ConvolutionInputGenerator.get_nodeattr_types(self))
        my_attrs.update(RTLBackend.get_nodeattr_types(self))
        return my_attrs

    def get_number_input_values(self):
        """Function to get the number of expected input values."""
        folded_ishape = self.get_folded_input_shape()
        num_input_elems = np.prod(folded_ishape[:-1])
        return num_input_elems

    def use_parallel_window_output(self):
        return self.get_nodeattr("parallel_window")

    def get_buffer_depth(self):
        """Returns total depth of the internal buffer, depending on
        implementation style."""
        ifm_ch = self.get_nodeattr("IFMChannels")
        k = self.get_nodeattr("ConvKernelDim")
        ifm_dim = self.get_nodeattr("IFMDim")
        stride = self.get_nodeattr("Stride")
        dilation = self.get_nodeattr("Dilation")
        simd = self.get_nodeattr("SIMD")

        k_h, k_w = k
        h, w = ifm_dim
        stride_h, stride_w = stride
        dilation_h, dilation_w = dilation
        mmv_in = 1
        mmv_out = 1
        channel_factor = int(ifm_ch / simd)
        impl_style = self.select_impl_style()
        if impl_style == "default":
            buffer_min_size = (
                (k_h - 1) * dilation_h * w + (k_w - 1) * dilation_w + 1
            ) * channel_factor
            # add additional buffer space in case of stride > 1
            # this minimizes cycle count as it allows an earlier pre-load of inputs
            buffer_depth = (
                buffer_min_size
                + max(
                    0,
                    ((stride_w - 1) - (int(mmv_out * k_h * k_w / mmv_in))) * channel_factor,
                )
                + max(
                    0,
                    ((stride_h - 1) * w - (int(mmv_out * k_h * k_w / mmv_in))) * channel_factor,
                )
            )
        elif impl_style == "parallel":
            buffer_min_size = (
                (k_h - 1) * dilation_h * w + (k_w - 1) * dilation_w
            ) * channel_factor + 1
            buffer_depth = buffer_min_size + 1
        return buffer_depth

    def get_exp_cycles(self):
        impl_style = self.select_impl_style()

        if impl_style == "parallel":
            exp_cycles = self.get_number_input_values() + 2
        elif impl_style == "default":
            simd = self.get_nodeattr("SIMD")
            ifm_ch = self.get_nodeattr("IFMChannels")
            k = self.get_nodeattr("ConvKernelDim")
            ifm_dim = self.get_nodeattr("IFMDim")
            ofm_dim = self.get_nodeattr("OFMDim")
            stride = self.get_nodeattr("Stride")
            dilation = self.get_nodeattr("Dilation")
            depthwise = self.get_nodeattr("depthwise")
            ifm_dim_h, ifm_dim_w = ifm_dim
            ofm_dim_h, ofm_dim_w = ofm_dim
            k_h, k_w = k
            stride_h, stride_w = stride
            dilation_h, dilation_w = dilation

            channel_factor = int(ifm_ch / simd)
            if ifm_dim_h == 1 or ifm_dim_w == 1:
                # 1D case
                (
                    ifm_ch,
                    [ifm_dim_h, ifm_dim_w],
                    [ofm_dim_h, ofm_dim_w],
                    [k_h, k_w],
                    [stride_h, stride_w],
                    [dilation_h, dilation_w],
                ) = self.get_1d_conv_attrs_normalized()

                if depthwise:
                    exp_cycles = (
                        +ofm_dim_w * k_w * channel_factor
                        + channel_factor * (k_w - 1) * (stride_w - 1)
                        - (k_w - 1)
                        + 2
                    )
                else:
                    exp_cycles = ofm_dim_w * k_w * channel_factor + 2
            else:
                # 2D case
                buffer_min_size = (
                    (k_h - 1) * dilation_h * ifm_dim_w + (k_w - 1) * dilation_w + 1
                ) * channel_factor
                cycles_write_block = ofm_dim_w * k_w * k_h * channel_factor
                cycles_read_block = stride_w * ifm_dim_w * channel_factor
                max_cycles = max(cycles_write_block, cycles_read_block)
                if depthwise:
                    max_cycles += ofm_dim_w * (stride_w - 1) * (channel_factor - 1)
                exp_cycles = buffer_min_size + ofm_dim_h * max_cycles
                if depthwise:
                    exp_cycles += (stride_h - 1) * ifm_dim_w * channel_factor

        return int(exp_cycles)

    def bram_estimation(self):
        simd = self.get_nodeattr("SIMD")
        ram_style = self.get_nodeattr("ram_style")
        impl_style = self.select_impl_style()
        [k_h, k_w] = self.get_nodeattr("ConvKernelDim")
        [ifm_dim_h, ifm_dim_w] = self.get_nodeattr("IFMDim")
        [dilation_h, dilation_w] = self.get_nodeattr("Dilation")

        if ram_style == "block" or ram_style == "auto":
            buffer_width = simd * self.get_input_datatype().bitwidth()
            if impl_style == "default":
                buffer_depth = self.get_buffer_depth()
                buffer_count = 1
            elif impl_style == "parallel":
                if ifm_dim_h == 1 or ifm_dim_w == 1:
                    return 0  # 1D case (no line buffers needed)
                kernel_width = (k_w - 1) * dilation_w + 1
                buffer_depth = (ifm_dim_w - kernel_width) + ifm_dim_w * (dilation_h - 1)
                buffer_count = k_h - 1

            # NOTE: Actual BRAM usage might be lower in some cases
            # due to imperfect modeling of Vivado behavior
            if buffer_depth <= 512:
                ram_width = 36
            elif buffer_depth <= 1024:
                ram_width = 18
            elif buffer_depth <= 2048:
                ram_width = 9
            elif buffer_depth <= 4096:
                ram_width = 4
            elif buffer_depth <= 8192:
                ram_width = 2
            else:
                ram_width = 1

            ram_cascade_depth = math.ceil(buffer_depth / 16384)
            ram_cascade_width = math.ceil(buffer_width / ram_width)
            cascade_savings = 0
            if buffer_depth > 16384:
                remainder_depth = buffer_depth % 16384
                if remainder_depth <= 512:
                    remainder_width = 36
                elif remainder_depth <= 1024:
                    remainder_width = 18
                elif remainder_depth <= 2048:
                    remainder_width = 9
                elif remainder_depth <= 4096:
                    remainder_width = 4
                elif remainder_depth <= 8192:
                    remainder_width = 2
                else:
                    remainder_width = 1

                remainder_cascade_width = math.ceil(buffer_width / remainder_width)
                cascade_savings = ram_cascade_width - remainder_cascade_width

            return int((ram_cascade_depth * ram_cascade_width - cascade_savings) * buffer_count)
        else:
            return 0

    def lut_estimation(self):
        simd = self.get_nodeattr("SIMD")
        ram_style = self.get_nodeattr("ram_style")
        buffer_width = simd * self.get_input_datatype().bitwidth()
        buffer_depth = self.get_buffer_depth()
        if ram_style == "distributed":
            ram_luts = int(buffer_width * math.ceil(buffer_depth / 38))
        else:
            ram_luts = 0
        return 300 + ram_luts

    def uram_estimation(self):
        simd = self.get_nodeattr("SIMD")
        ram_style = self.get_nodeattr("ram_style")
        impl_style = self.select_impl_style()
        [k_h, k_w] = self.get_nodeattr("ConvKernelDim")
        [ifm_dim_h, ifm_dim_w] = self.get_nodeattr("IFMDim")
        [dilation_h, dilation_w] = self.get_nodeattr("Dilation")

        if ram_style == "ultra":
            buffer_width = simd * self.get_input_datatype().bitwidth()
            if impl_style == "default":
                buffer_depth = self.get_buffer_depth()
                buffer_count = 1
            elif impl_style == "parallel":
                if ifm_dim_h == 1 or ifm_dim_w == 1:
                    return 0  # 1D case (no line buffers needed)
                kernel_width = (k_w - 1) * dilation_w + 1
                buffer_depth = (ifm_dim_w - kernel_width) + ifm_dim_w * (dilation_h - 1)
                buffer_count = k_h - 1

            ram_depth = 4096
            ram_width = 72
            ram_cascade_depth = math.ceil(buffer_depth / ram_depth)
            ram_cascade_width = math.ceil(buffer_width / ram_width)
            return int(ram_cascade_depth * ram_cascade_width * buffer_count)
        else:
            return 0

    def execute_node(self, context, graph):
        mode = self.get_nodeattr("exec_mode")

        if mode == "cppsim":
            ConvolutionInputGenerator.execute_node(self, context, graph)
            # if depthwise = 1
            # interleave channels such that cppsim of ConvolutionInputGenerator_rtl
            # has a notion of SIMD parallelism. Subsequent VVAU_{hls/rtl} expects
            # the channels to be interleaved (i.e. to match their PE parallelism).
            if self.get_nodeattr("depthwise"):
                node = self.onnx_node
                im2col_out = context[node.output[0]]
                simd = getCustomOp(node).get_nodeattr("SIMD")
                ofm_h, ofm_w = getCustomOp(node).get_nodeattr("OFMDim")
                k_h, k_w = getCustomOp(node).get_nodeattr("ConvKernelDim")
                ifm_ch = getCustomOp(node).get_nodeattr("IFMChannels")
                im2col_out = im2col_out.reshape(1, ofm_h, ofm_w, k_h * k_w, ifm_ch // simd, simd)
                im2col_out = im2col_out.transpose(0, 1, 2, 4, 3, 5)
                im2col_out = im2col_out.reshape(1, ofm_h, ofm_w, ifm_ch * k_h * k_w)
                context[node.output[0]] = im2col_out
        elif mode == "rtlsim":
            RTLBackend.execute_node(self, context, graph)

    def prepare_codegen_default(self):
        """Fills code generation dict for the default implementation style by computing
        the incremental addressing scheme for the circular buffer."""
        if self.get_nodeattr("dynamic_mode"):
            template_select = "swg/swg_template_default_dynamic.sv"
        else:
            template_select = "swg/swg_template_default.sv"
        template_path = os.path.join(os.environ["FINN_RTLLIB"], template_select)
        code_gen_dict = {}

        ifm_ch = self.get_nodeattr("IFMChannels")
        k = self.get_nodeattr("ConvKernelDim")
        ifm_dim = self.get_nodeattr("IFMDim")
        stride = self.get_nodeattr("Stride")
        dilation = self.get_nodeattr("Dilation")
        depthwise = self.get_nodeattr("depthwise")
        simd = self.get_nodeattr("SIMD")

        k_h, k_w = k
        h, w = ifm_dim
        pad = [0, 0, 0, 0]  # padding happens in separate padding node for now
        stride_h, stride_w = stride
        dilation_h, dilation_w = dilation
        pad_h = pad[0] + pad[2]
        pad_w = pad[1] + pad[3]
        out_dim_h = im2col.compute_conv_output_dim(h, k_h, stride_h, pad_h, dilation_h)
        out_dim_w = im2col.compute_conv_output_dim(w, k_w, stride_w, pad_w, dilation_w)
        mmv_in = 1
        mmv_out = 1
        channel_factor = int(ifm_ch / simd)

        # compute minimal buffer length (assuming it holds 1 complete window)
        buffer_min_size = ((k_h - 1) * dilation_h * w + (k_w - 1) * dilation_w + 1) * channel_factor

        buffer_actual_size = self.get_buffer_depth()
        code_gen_dict["$BUF_ELEM_TOTAL$"] = [str(buffer_actual_size)]

        # compute some intermediate values, e.g., kernel "width" = k_w incl. dilation
        # or cols/rows that are skipped due to imperfect stride<->dim combination
        kernel_width = (k_w - 1) * dilation_w + 1
        kernel_height = (k_h - 1) * dilation_h + 1
        skip_columns = w % (kernel_width + (out_dim_w - 1) * stride_w)
        skip_rows = h % (kernel_height + (out_dim_h - 1) * stride_h)

        # compute address increment values for 5-loop nest
        addr_incr_end_simd = 1
        addr_incr_end_window_elem = (dilation_w - 1) * channel_factor + 1
        addr_incr_end_window_row = (
            ((w - kernel_width) * channel_factor)  # remaining line
            + ((dilation_h - 1) * w * channel_factor)  # skip lines
            + 1  # wrap-around of minimally sized buffer
        )
        addr_incr_end_window = -buffer_min_size + stride_w * channel_factor + 1
        addr_incr_end_row = (
            -buffer_min_size
            + ((skip_columns + kernel_width) * channel_factor)  # remaining line
            + ((stride_h - 1) * w * channel_factor)  # skip lines
            + 1
        )

        # re-use same controller structure -> re-assign address increments
        if depthwise:
            addr_incr_end_window_elem = dilation_w * channel_factor
            addr_incr_end_window_row = (
                channel_factor
                + (w - kernel_width) * channel_factor
                + (dilation_h - 1) * w * channel_factor
            )
            addr_incr_end_simd = -buffer_min_size + (channel_factor + 1)

        # sanity check for wrap logic
        assert not (
            abs(addr_incr_end_window) > buffer_actual_size
        ), "ERROR: W increment > buffer size, try setting parallel_window=1"
        assert not (
            abs(addr_incr_end_row) > buffer_actual_size
        ), "ERROR: H increment > buffer size, try setting parallel_window=1"

        # set certain threshold indices to detect when reading/writing finishes
        code_gen_dict["$LAST_READ_ELEM$"] = [str(h * w * channel_factor - 1)]
        code_gen_dict["$LAST_WRITE_ELEM$"] = [
            str(((h - skip_rows - 1) * w + (w - skip_columns)) * channel_factor - 1)
        ]

        # default controller loop structure: # iterations (counters) map directly
        loop_h_iterations = out_dim_h
        loop_w_iterations = out_dim_w
        loop_kh_iterations = k_h
        loop_kw_iterations = k_w
        loop_simd_iterations = channel_factor

        if depthwise and channel_factor > 1:
            # re-arrange existing controller loop structure for depthwise convolutions
            loop_kh_iterations = channel_factor
            loop_kw_iterations = k_h
            loop_simd_iterations = k_w
            addr_incr_end_simd_ = addr_incr_end_simd
            addr_incr_end_simd = addr_incr_end_window_elem
            addr_incr_end_window_elem = addr_incr_end_window_row
            addr_incr_end_window_row = addr_incr_end_simd_
            elem_per_window = k_h * k_w

            tail_incr_w = addr_incr_end_window + buffer_min_size - channel_factor
            tail_incr_h = addr_incr_end_row + buffer_min_size - channel_factor
            tail_incr_last_window = buffer_min_size - 1
            code_gen_dict["$IS_DEPTHWISE$"] = ["1"]
        else:
            # depthwise output format is equivalent to non-depthwise if SIMD=C
            elem_per_window = k_h * k_w * channel_factor

            tail_incr_w = addr_incr_end_window + buffer_min_size - 1
            tail_incr_h = addr_incr_end_row + buffer_min_size - 1
            tail_incr_last_window = buffer_min_size - 1
            code_gen_dict["$IS_DEPTHWISE$"] = ["0"]

        # support SIMD = IFMChannels and k_w = 1 cases
        # for k = [k_h, k_w] = [1, k_w], no adjustment is needed
        # for k = [k_h, k_w] = [1, 1], do not use this impl. style (mmv_out=K=1)
        # innermost loop is executed at least once -> adjust if needed
        if loop_simd_iterations == 1:
            # skip innermost SIMD loop completely
            if loop_kw_iterations == 1:
                # skip innermost KW loop completely
                code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_KH"]
                loop_kh_iterations -= 1  # -1 because state is initial state
            else:
                code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_KW"]
                loop_kw_iterations -= 1  # -1 because state is initial state
        else:
            code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_SIMD"]
            loop_simd_iterations -= 1  # -1 because state is initial state

        cntr_bitwidth = math.ceil(
            math.log2(
                max(
                    loop_h_iterations - 2 + 1,
                    loop_w_iterations - 2 + 1,
                    loop_kh_iterations - 2 + 1,
                    loop_kw_iterations - 2 + 1,
                    loop_simd_iterations - 2 + 1,
                )
            )
        )
        code_gen_dict["$CNTR_BITWIDTH$"] = [str(cntr_bitwidth)]
        code_gen_dict["$LOOP_H_ITERATIONS$"] = [str(loop_h_iterations - 2)]
        code_gen_dict["$LOOP_W_ITERATIONS$"] = [str(loop_w_iterations - 2)]
        code_gen_dict["$LOOP_KH_ITERATIONS$"] = [str(loop_kh_iterations - 2)]
        code_gen_dict["$LOOP_KW_ITERATIONS$"] = [str(loop_kw_iterations - 2)]
        code_gen_dict["$LOOP_SIMD_ITERATIONS$"] = [str(loop_simd_iterations - 2)]

        incr_bitwidth = 1 + math.ceil(
            math.log2(
                max(
                    abs(addr_incr_end_simd) + 1,
                    abs(addr_incr_end_window_elem) + 1,
                    abs(addr_incr_end_window_row) + 1,
                    abs(addr_incr_end_window) + 1,
                    abs(addr_incr_end_row) + 1,
                    abs(tail_incr_w) + 1,
                    abs(tail_incr_h) + 1,
                    abs(tail_incr_last_window) + 1,
                )
            )
        )
        code_gen_dict["$INCR_BITWIDTH$"] = [str(incr_bitwidth)]
        code_gen_dict["$HEAD_INCR_SIMD$"] = [str(addr_incr_end_simd)]
        code_gen_dict["$HEAD_INCR_KW$"] = [str(addr_incr_end_window_elem)]
        code_gen_dict["$HEAD_INCR_KH$"] = [str(addr_incr_end_window_row)]
        code_gen_dict["$HEAD_INCR_W$"] = [str(addr_incr_end_window)]
        code_gen_dict["$HEAD_INCR_H$"] = [str(addr_incr_end_row)]
        code_gen_dict["$TAIL_INCR_W$"] = [str(tail_incr_w)]
        code_gen_dict["$TAIL_INCR_H$"] = [str(tail_incr_h)]
        code_gen_dict["$TAIL_INCR_LAST$"] = [str(tail_incr_last_window)]

        code_gen_dict["$ELEM_PER_WINDOW$"] = [str(elem_per_window)]
        code_gen_dict["$SIMD$"] = [str(simd)]
        code_gen_dict["$MMV_IN$"] = [str(mmv_in)]
        code_gen_dict["$MMV_OUT$"] = [str(mmv_out)]

        return template_path, code_gen_dict

    def prepare_codegen_parallel(self):
        """Fills code generation dict for the parallel implementation style by computing
        the loop controller configuration and partitioning the fixed buffer into
        shift-registers (for parallel read access) and line buffers (for efficient
        LUTRAM/BRAM/URAM implementation)."""
        template_path = os.path.join(os.environ["FINN_RTLLIB"], "swg/swg_template_parallel.sv")
        code_gen_dict = {}

        ifm_ch = self.get_nodeattr("IFMChannels")
        k = self.get_nodeattr("ConvKernelDim")
        ifm_dim = self.get_nodeattr("IFMDim")
        stride = self.get_nodeattr("Stride")
        dilation = self.get_nodeattr("Dilation")
        simd = self.get_nodeattr("SIMD")
        M = self.get_nodeattr("M")

        k_h, k_w = k
        h, w = ifm_dim
        pad = [0, 0, 0, 0]  # padding happens in separate padding node for now
        stride_h, stride_w = stride
        dilation_h, dilation_w = dilation
        pad_h = pad[0] + pad[2]
        pad_w = pad[1] + pad[3]
        out_dim_h = im2col.compute_conv_output_dim(h, k_h, stride_h, pad_h, dilation_h)
        out_dim_w = im2col.compute_conv_output_dim(w, k_w, stride_w, pad_w, dilation_w)
        mmv_in = M * 1
        mmv_out = M * k_h * k_w
        channel_factor = int(ifm_ch / simd)

        # compute minimal buffer length (assuming it holds 1 complete window)
        buffer_min_size = ((k_h - 1) * dilation_h * w + (k_w - 1) * dilation_w) * channel_factor + 1

        buffer_actual_size = self.get_buffer_depth()
        code_gen_dict["$BUF_ELEM_TOTAL$"] = [str(buffer_actual_size)]

        # compute some intermediate values, e.g., kernel "width" = k_w incl. dilation
        # or cols/rows that are skipped due to imperfect stride<->dim combination
        kernel_width = (k_w - 1) * dilation_w + 1
        kernel_height = (k_h - 1) * dilation_h + 1
        skip_columns = w % (kernel_width + (out_dim_w - 1) * stride_w)
        skip_rows = h % (kernel_height + (out_dim_h - 1) * stride_h)

        # set certain threshold indices to detect when reading/writing finishes
        code_gen_dict["$LAST_READ_ELEM$"] = [str(h * w * channel_factor - 1)]
        code_gen_dict["$LAST_WRITE_ELEM$"] = [
            str(((h - skip_rows - 1) * w + (w - skip_columns)) * channel_factor - 1)
        ]

        # re-use default controller loop structure
        loop_h_iterations = out_dim_h
        loop_w_iterations = out_dim_w
        loop_kh_iterations = channel_factor
        loop_kw_iterations = 1
        loop_simd_iterations = 1

        if loop_kh_iterations == 1:
            if loop_w_iterations == 1:
                code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_H"]
                loop_h_iterations -= 1  # -1 because state is initial state
            else:
                code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_W"]
                loop_w_iterations -= 1  # -1 because state is initial state
        else:
            code_gen_dict["$INNERMOST_STATE$"] = ["STATE_LOOP_KH"]
            loop_kh_iterations -= 1  # -1 because state is initial state

        # set head address increment values
        addr_incr_end_simd = 1
        addr_incr_end_window_elem = 1
        addr_incr_end_window_row = 1
        addr_incr_end_window = (stride_w - 1) * channel_factor + 1
        addr_incr_end_row = ((skip_columns + (kernel_width - 1)) * channel_factor + 1) + (
            (stride_h - 1) * w * channel_factor
        )

        # add init value for CURRENT_ELEM counter = last elem of first window
        code_gen_dict["$FIRST_WRITE_ELEM$"] = [str(buffer_min_size - 1)]

        cntr_bitwidth = math.ceil(
            math.log2(
                max(
                    loop_h_iterations - 2 + 1,
                    loop_w_iterations - 2 + 1,
                    loop_kh_iterations - 2 + 1,
                    loop_kw_iterations - 2 + 1,
                    loop_simd_iterations - 2 + 1,
                )
            )
        )
        code_gen_dict["$CNTR_BITWIDTH$"] = [str(cntr_bitwidth)]
        code_gen_dict["$LOOP_H_ITERATIONS$"] = [str(loop_h_iterations - 2)]
        code_gen_dict["$LOOP_W_ITERATIONS$"] = [str(loop_w_iterations - 2)]
        code_gen_dict["$LOOP_KH_ITERATIONS$"] = [str(loop_kh_iterations - 2)]
        code_gen_dict["$LOOP_KW_ITERATIONS$"] = [str(loop_kw_iterations - 2)]
        code_gen_dict["$LOOP_SIMD_ITERATIONS$"] = [str(loop_simd_iterations - 2)]

        incr_bitwidth = 1 + math.ceil(
            math.log2(
                max(
                    abs(addr_incr_end_simd) + 1,
                    abs(addr_incr_end_window_elem) + 1,
                    abs(addr_incr_end_window_row) + 1,
                    abs(addr_incr_end_window) + 1,
                    abs(addr_incr_end_row) + 1,
                )
            )
        )
        code_gen_dict["$INCR_BITWIDTH$"] = [str(incr_bitwidth)]
        code_gen_dict["$HEAD_INCR_SIMD$"] = [str(addr_incr_end_simd)]
        code_gen_dict["$HEAD_INCR_KW$"] = [str(addr_incr_end_window_elem)]
        code_gen_dict["$HEAD_INCR_KH$"] = [str(addr_incr_end_window_row)]
        code_gen_dict["$HEAD_INCR_W$"] = [str(addr_incr_end_window)]
        code_gen_dict["$HEAD_INCR_H$"] = [str(addr_incr_end_row)]
        # not used, set to zero:
        code_gen_dict["$TAIL_INCR_W$"] = ["0"]
        code_gen_dict["$TAIL_INCR_H$"] = ["0"]
        code_gen_dict["$TAIL_INCR_LAST$"] = ["0"]
        code_gen_dict["$IS_DEPTHWISE$"] = ["0"]

        code_gen_dict["$SIMD$"] = [str(simd)]
        code_gen_dict["$MMV_IN$"] = [str(mmv_in)]
        code_gen_dict["$MMV_OUT$"] = [str(mmv_out)]

        # prepare buffer partitioning into "reg_fifos" and "bram_fifos"
        # use normalized ([H,W]=[1,W]) dimensions for 1D case
        (
            ifm_ch,
            [ifm_dim_h, ifm_dim_w],
            [ofm_dim_h, ofm_dim_w],
            [k_h, k_w],
            [stride_h, stride_w],
            [dilation_h, dilation_w],
        ) = self.get_1d_conv_attrs_normalized()

        reg_fifos = []
        bram_fifos_depth = []

        px_idx = 0
        for ky in range(k_h):
            reg_fifo = []
            for kx in range(k_w):
                for c in range(channel_factor):
                    if c < (channel_factor - 1):
                        if not (ky == 0 and kx == 0):
                            reg_fifo.append(-1)
                            px_idx += 1
                    else:
                        reg_fifo.append(px_idx)
                        px_idx += 1
                if kx < (k_w - 1):
                    reg_fifo.extend([-1] * ((dilation_w - 1) * channel_factor))
                    px_idx += (dilation_w - 1) * channel_factor
            reg_fifos.append(reg_fifo)

            if ky < (k_h - 1):
                line_buffer_len = ((w - kernel_width) + w * (dilation_h - 1)) * channel_factor
                bram_fifos_depth.append(line_buffer_len)
                px_idx += line_buffer_len

        code_gen_dict["$GENERATE_REG_FIFOS$"] = []
        for i, reg_fifo in enumerate(reg_fifos):
            code_gen_dict["$GENERATE_REG_FIFOS$"].append(
                """
                wire [IN_WIDTH-1:0] reg_fifo_{id}_in;
                wire [IN_WIDTH-1:0] reg_fifo_{id}_out;
                wire [IN_WIDTH*{len}-1:0] reg_fifo_{id};
                swg_reg_buffer
                #(
                .WIDTH(IN_WIDTH),
                .DEPTH({len})
                )
                reg_buffer_inst_{id}
                (
                    .clk(clk),
                    .shift_enable(shift_enable),
                    .shift_in(reg_fifo_{id}_in),
                    .shift_out(reg_fifo_{id}_out),
                    .data_out(reg_fifo_{id})
                );""".format(
                    id=i,
                    len=len(reg_fifo),
                )
            )

        code_gen_dict["$GENERATE_BRAM_FIFOS$"] = []
        for i, bram_fifo_depth in enumerate(bram_fifos_depth):
            code_gen_dict["$GENERATE_BRAM_FIFOS$"].append(
                """
                wire [IN_WIDTH-1:0] bram_fifo_{id}_in;
                wire [IN_WIDTH-1:0] bram_fifo_{id}_out;
                swg_ram_buffer
                #(
                .WIDTH(IN_WIDTH),
                .DEPTH({len}),
                .RAM_STYLE("{ram_style}")
                )
                ram_buffer_inst_{id}
                (
                    .clk(clk),
                    .rst_n(rst_n),
                    .shift_enable(shift_enable),
                    .shift_in(bram_fifo_{id}_in),
                    .shift_out(bram_fifo_{id}_out)
                );""".format(
                    id=i,
                    len=bram_fifo_depth,
                    ram_style=self.get_nodeattr("ram_style"),
                )
            )

        code_gen_dict["$GENERATE_OUTPUT_MAPPING$"] = []
        out_idx = mmv_out - 1
        for fifo_id, reg_fifo in enumerate(reg_fifos):
            for fifo_idx, access_idx in enumerate(reg_fifo):
                if access_idx != -1:
                    code_gen_dict["$GENERATE_OUTPUT_MAPPING$"].append(
                        """assign data_out[OUT_ELEM_WIDTH*{out_idx}+:OUT_ELEM_WIDTH]
                        = reg_fifo_{fifo_id}[{access_idx}*{mmv}*OUT_ELEM_WIDTH+
                        OUT_ELEM_WIDTH*{mmv_idx}+:OUT_ELEM_WIDTH];""".format(
                            out_idx=out_idx,
                            fifo_id=fifo_id,
                            access_idx=len(reg_fifo) - 1 - int((max(reg_fifo) - access_idx) / M),
                            mmv_idx=(max(reg_fifo) - access_idx) % M,
                            mmv=M,
                        )
                    )
                    # reversal: out_idx=0 -> oldest buffer element -> highest access_idx
                    out_idx = out_idx - 1
        assert out_idx == -1, "ERROR: Not all output vector elements connected"

        code_gen_dict["$GENERATE_BUFFER_CONNECTION$"] = []
        for i in range(len(reg_fifos)):
            if i == 0:
                # first FIFO containing newest elements -> input comes from input reg
                code_gen_dict["$GENERATE_BUFFER_CONNECTION$"].append(
                    """assign reg_fifo_{fifo_id}_in = data_in;""".format(
                        fifo_id=i,
                    )
                )
            else:
                # other REG FIFOs -> input comes from connected BRAM FIFO (line buffer)
                input_fifo_id = i - 1
                code_gen_dict["$GENERATE_BUFFER_CONNECTION$"].append(
                    """assign reg_fifo_{fifo_id}_in = bram_fifo_{input_fifo_id}_out;
                    """.format(
                        fifo_id=i, input_fifo_id=input_fifo_id
                    )
                )
        for i in range(len(bram_fifos_depth)):
            input_fifo_id = i
            code_gen_dict["$GENERATE_BUFFER_CONNECTION$"].append(
                """assign bram_fifo_{fifo_id}_in = reg_fifo_{input_fifo_id}_out;
                """.format(
                    fifo_id=i, input_fifo_id=input_fifo_id
                )
            )

        return template_path, code_gen_dict

    def select_impl_style(self):
        """Selects implementation style based on folding configuration."""
        simd = self.get_nodeattr("SIMD")
        M = self.get_nodeattr("M")
        depthwise = self.get_nodeattr("depthwise")
        ifm_ch = self.get_nodeattr("IFMChannels")
        ifm_dim = self.get_nodeattr("IFMDim")
        stride = self.get_nodeattr("Stride")
        dilation = self.get_nodeattr("Dilation")
        k = self.get_nodeattr("ConvKernelDim")
        ifm_dim_h, ifm_dim_w = ifm_dim
        stride_h, stride_w = stride
        dilation_h, dilation_w = dilation
        k_h, k_w = k
        kernel_width = (k_w - 1) * dilation_w + 1  # incl. dilation
        kernel_height = (k_h - 1) * dilation_h + 1  # incl. dilation

        # check for valid configuration
        assert (
            kernel_height <= ifm_dim_h
            and kernel_width <= ifm_dim_w
            and stride_h <= ifm_dim_h
            and stride_w <= ifm_dim_w
        ), "Illegal conv configuration: kernel or stride > FM dimension"

        # init folding config
        if self.get_nodeattr("parallel_window"):
            # mmv_in = M * 1
            mmv_out = M * k_h * k_w
        else:
            # mmv_in = 1
            mmv_out = 1
            assert ifm_ch % simd == 0, "Constraint violated: SIMD must divide IFMChannels"

        # choose implementation style
        if mmv_out > 1 or (k_h == 1 and k_w == 1):
            impl_style = "parallel"
            if depthwise or (k_h == 1 and k_w == 1):
                # allow SIMD < IFM_CH in depthwise mode (VVAU supports the resulting data layout)
                # also allowed for 1x1 kernel since depthwise and non-depthwise are equivalent
                assert ifm_ch % simd == 0, "Constraint violated: SIMD must divide IFMChannels"
            else:
                assert ifm_ch == simd, "Constraint violated: SIMD must be equal to IFMChannels"
        else:
            impl_style = "default"

        return impl_style

    def generate_hdl(self, model, fpgapart, clk):
        """Generates HDL code and wrapper for the IP, depending on required
        implementation style."""
        impl_style = self.select_impl_style()

        # prepare code generation by filling out dictionaries
        if impl_style == "default":
            template_path, code_gen_dict = self.prepare_codegen_default()
        elif impl_style == "parallel":
            template_path, code_gen_dict = self.prepare_codegen_parallel()
            if self.get_nodeattr("dynamic_mode"):
                raise Exception("Dynamic mode is not compatible with parallel_window")
        else:
            raise Exception("Requested impl. style not implemented")

        # add general parameters to dictionary
        code_gen_dict["$TOP_MODULE_NAME$"] = [self.get_verilog_top_module_name()]
        # save top module name so we can refer to it after this node has been renamed
        # (e.g. by GiveUniqueNodeNames(prefix) during MakeZynqProject)
        self.set_nodeattr("gen_top_module", self.get_verilog_top_module_name())
        code_gen_dict["$BIT_WIDTH$"] = [str(self.get_input_datatype().bitwidth())]
        code_gen_dict["$IN_WIDTH_PADDED$"] = [
            str(roundup_to_integer_multiple(self.get_instream_width(), 8))
        ]
        code_gen_dict["$OUT_WIDTH_PADDED$"] = [
            str(roundup_to_integer_multiple(self.get_outstream_width(), 8))
        ]
        ram_style = self.get_nodeattr("ram_style")
        code_gen_dict["$RAM_STYLE$"] = ['"{}"'.format(ram_style)]

        # apply code generation to templates
        code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")
        with open(template_path, "r") as f:
            template = f.read()
        if self.get_nodeattr("dynamic_mode"):
            template_select = "swg/swg_template_wrapper_dynamic.v"
        else:
            template_select = "swg/swg_template_wrapper.v"
        with open(os.path.join(os.environ["FINN_RTLLIB"], template_select), "r") as f:
            template_wrapper = f.read()
        with open(os.path.join(os.environ["FINN_RTLLIB"], "swg/swg_template_axilite.v"), "r") as f:
            template_axilite = f.read()
        for key in code_gen_dict:
            # transform list into long string separated by '\n'
            code_gen_line = "\n".join(code_gen_dict[key])
            template = template.replace(key, code_gen_line)
            template_wrapper = template_wrapper.replace(key, code_gen_line)
            template_axilite = template_axilite.replace(key, code_gen_line)
        with open(
            os.path.join(code_gen_dir, self.get_nodeattr("gen_top_module") + "_impl.sv"),
            "w",
        ) as f:
            f.write(template)
        with open(
            os.path.join(code_gen_dir, self.get_nodeattr("gen_top_module") + "_wrapper.v"),
            "w",
        ) as f:
            f.write(template_wrapper)

        # AXI-Lite reg. file component is only needed for dynamic mode
        if self.get_nodeattr("dynamic_mode"):
            with open(
                os.path.join(code_gen_dir, self.get_nodeattr("gen_top_module") + "_axilite.v"),
                "w",
            ) as f:
                f.write(template_axilite)

        # Copy static source file for common core components
        shutil.copy2(os.path.join(os.environ["FINN_RTLLIB"], "swg/swg_common.sv"), code_gen_dir)
        shutil.copy2(os.path.join(os.environ["FINN_RTLLIB"], "swg/swg_pkg.sv"), code_gen_dir)

        # set ipgen_path and ip_path so that HLS-Synth transformation
        # and stich_ip transformation do not complain
        self.set_nodeattr("ipgen_path", code_gen_dir)
        self.set_nodeattr("ip_path", code_gen_dir)

    def get_rtl_file_list(self, abspath=False):
        if abspath:
            code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen") + "/"
            rtllib_dir = os.path.join(os.environ["FINN_RTLLIB"], "swg/")
        else:
            code_gen_dir = ""
            rtllib_dir = ""
        verilog_files = [
            rtllib_dir + "swg_pkg.sv",
            code_gen_dir + self.get_nodeattr("gen_top_module") + "_wrapper.v",
            code_gen_dir + self.get_nodeattr("gen_top_module") + "_impl.sv",
            rtllib_dir + "swg_common.sv",
        ]
        if self.get_nodeattr("dynamic_mode"):
            verilog_files.append(code_gen_dir + self.get_nodeattr("gen_top_module") + "_axilite.v")

        return verilog_files

    def code_generation_ipi(self):
        """Constructs and returns the TCL for node instantiation in Vivado IPI."""
        code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")

        sourcefiles = [
            "swg_pkg.sv",
            self.get_nodeattr("gen_top_module") + "_wrapper.v",
            self.get_nodeattr("gen_top_module") + "_impl.sv",
            "swg_common.sv",
        ]

        if self.get_nodeattr("dynamic_mode"):
            sourcefiles += [self.get_nodeattr("gen_top_module") + "_axilite.v"]

        sourcefiles = [os.path.join(code_gen_dir, f) for f in sourcefiles]

        cmd = []
        for f in sourcefiles:
            cmd += ["add_files -norecurse %s" % (f)]
        cmd += [
            "create_bd_cell -type module -reference %s %s"
            % (self.get_nodeattr("gen_top_module"), self.onnx_node.name)
        ]
        return cmd

    def get_verilog_top_module_intf_names(self):
        # Overload default HLSCustomOp implementation to add axilite control IF
        """Return a dict of names of input and output interfaces.
        The keys reflect the protocols each interface implements:
        'clk', 'rst', 'm_axis', 's_axis', 'aximm', 'axilite'.
        Values are lists of tuples (axis, aximm) or names (axilite):
        'axis' tuples correspond to the list of node inputs in order,
        each tuple is (interface_name, interface_width_bits).
        axilite always assumed to be 32 bits and is not tuple (name only).
        Each block must have at most one aximm and one axilite."""
        intf_names = super().get_verilog_top_module_intf_names()
        if self.get_nodeattr("dynamic_mode"):
            intf_names["axilite"] = ["s_axilite"]
        return intf_names

    def get_dynamic_config(self, ifm_dim=None, stride=None, dilation=None):
        """Returns a configuration dict to re-configure FM dimension during
        runtime. Stride and dilation can also be changed. Certain restrictions
        apply (e.g. component must be synthesized for largest buffer size)."""
        # NOTE: For better driver integration, this functionality could be packaged
        # as a standalone function in the future
        if self.select_impl_style() != "default":
            raise Exception("Impl. style is incompatible with dynamic mode")

        if ifm_dim is None:
            ifm_dim = self.get_nodeattr("IFMDim")
        k = self.get_nodeattr("ConvKernelDim")
        if stride is None:
            stride = self.get_nodeattr("Stride")
        if dilation is None:
            dilation = self.get_nodeattr("Dilation")

        k_h, k_w = k
        stride_h, stride_w = stride
        dilation_h, dilation_w = dilation
        ifm_dim_h, ifm_dim_w = ifm_dim
        ofm_dim_h = compute_conv_output_dim(ifm_dim_h, k_h, stride_h, 0, dilation_h)
        ofm_dim_w = compute_conv_output_dim(ifm_dim_w, k_w, stride_w, 0, dilation_w)
        ofm_dim = [ofm_dim_h, ofm_dim_w]

        # update attributes and perform sanity check
        original_buffer_depth = self.get_buffer_depth()
        self.set_nodeattr("IFMDim", ifm_dim)
        self.set_nodeattr("OFMDim", ofm_dim)
        self.set_nodeattr("Stride", stride)
        self.set_nodeattr("Dilation", dilation)
        assert (
            self.get_buffer_depth() <= original_buffer_depth
        ), """Error: requested
            dynamic configuration does not fit in generated buffer implementation."""

        # (re-)call codegen and extract new values
        # each setting is mapped to an axi-lite register address
        template_path, code_gen_dict = self.prepare_codegen_default()
        config = {
            "cfg_wren": (0 * 4, 1),
            "cfg_cntr_simd": (1 * 4, int(code_gen_dict["$LOOP_SIMD_ITERATIONS$"][0])),
            "cfg_cntr_kw": (2 * 4, int(code_gen_dict["$LOOP_KW_ITERATIONS$"][0])),
            "cfg_cntr_kh": (3 * 4, int(code_gen_dict["$LOOP_KH_ITERATIONS$"][0])),
            "cfg_cntr_w": (4 * 4, int(code_gen_dict["$LOOP_W_ITERATIONS$"][0])),
            "cfg_cntr_h": (5 * 4, int(code_gen_dict["$LOOP_H_ITERATIONS$"][0])),
            "cfg_incr_head_simd": (6 * 4, int(code_gen_dict["$HEAD_INCR_SIMD$"][0])),
            "cfg_incr_head_kw": (7 * 4, int(code_gen_dict["$HEAD_INCR_KW$"][0])),
            "cfg_incr_head_kh": (8 * 4, int(code_gen_dict["$HEAD_INCR_KH$"][0])),
            "cfg_incr_head_w": (9 * 4, int(code_gen_dict["$HEAD_INCR_W$"][0])),
            "cfg_incr_head_h": (10 * 4, int(code_gen_dict["$HEAD_INCR_H$"][0])),
            "cfg_incr_tail_w": (11 * 4, int(code_gen_dict["$TAIL_INCR_W$"][0])),
            "cfg_incr_tail_h": (12 * 4, int(code_gen_dict["$TAIL_INCR_H$"][0])),
            "cfg_incr_tail_last": (13 * 4, int(code_gen_dict["$TAIL_INCR_LAST$"][0])),
            "cfg_last_read": (14 * 4, int(code_gen_dict["$LAST_READ_ELEM$"][0])),
            "cfg_last_write": (15 * 4, int(code_gen_dict["$LAST_WRITE_ELEM$"][0])),
        }
        return config



"""Transformations for generating and simulating instrumentation IP."""

import numpy as np
import os
import subprocess
from pathlib import Path
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation

from finn.custom_op.fpgadataflow.templates import ipgentcl_template
from finn.util.basic import make_build_dir
from finn.util.deps import get_deps_path
from finn.util.hls import CallHLS


# TODO: duplicate function from make_zynq_proj.py
def collect_ip_dirs(model, ipstitch_path):
    """Collect list of all IP directories required by the design."""
    ip_dirs = []
    need_memstreamer = False
    for node in model.graph.node:
        node_inst = getCustomOp(node)
        ip_dir_value = node_inst.get_nodeattr("ip_path")
        assert os.path.isdir(
            ip_dir_value
        ), """The directory that should
        contain the generated ip blocks doesn't exist."""
        ip_dirs += [ip_dir_value]
        if node.op_type.startswith("MVAU") or node.op_type == "Thresholding_hls":
            if node_inst.get_nodeattr("mem_mode") == "internal_decoupled":
                need_memstreamer = True
    ip_dirs += [ipstitch_path + "/ip"]
    if need_memstreamer:
        # add RTL streamer IP
        ip_dirs.append("$::env(FINN_RTLLIB)/memstream")
    return ip_dirs


class GenerateInstrumentationIP(Transformation):
    """Generate instrumentation IP for performance monitoring."""

    def __init__(
        self,
        fpga_part,
        clk_period_ns,
        format="ip",  # "ip" for Vivado (Zynq) or "xo" for Vitis (Alveo/Versal)
    ):
        """Initialize instrumentation IP generation with FPGA part and clock settings."""
        super().__init__()
        self.fpga_part = fpga_part
        self.clk_period_ns = clk_period_ns
        self.format = format

    def apply(self, model):
        """Generate instrumentation IP core."""
        # Create directory for code-gen and HLS of instrumentation IP
        wrapper_output_dir = make_build_dir(prefix="code_gen_ipgen_Instrumentation_")
        model.set_metadata_prop("instrumentation_ipgen", wrapper_output_dir)

        # conservative max for pending feature maps: number of layers
        pending = len(model.graph.node)
        # query the parallelism-dependent folded input shape from the
        # node consuming the graph input
        inp_name = model.graph.input[0].name
        inp_node = getCustomOp(model.find_consumer(inp_name))
        inp_shape_folded = list(inp_node.get_folded_input_shape())
        inp_stream_width = inp_node.get_instream_width_padded()
        # number of beats per input is given by product of folded input
        # shape except the last dim (which is the stream width)
        ilen = np.prod(inp_shape_folded[:-1])
        ti = "ap_uint<%d>" % inp_stream_width
        # perform the same for the output
        out_name = model.graph.output[0].name
        out_node = getCustomOp(model.find_producer(out_name))
        out_shape_folded = list(out_node.get_folded_output_shape())
        out_stream_width = out_node.get_outstream_width_padded()
        olen = np.prod(out_shape_folded[:-1])
        to = "ap_uint<%d>" % out_stream_width
        ko = out_shape_folded[-1]
        # fill out instrumentation wrapper template
        with open(
            os.path.join(os.environ["FINN_CUSTOM_HLS"], "instrumentation.template.cpp"), "r"
        ) as f:
            instrwrp_cpp = f.read()
        instrwrp_cpp = instrwrp_cpp.replace("@PENDING@", str(pending))
        instrwrp_cpp = instrwrp_cpp.replace("@ILEN@", str(ilen))
        instrwrp_cpp = instrwrp_cpp.replace("@OLEN@", str(olen))
        instrwrp_cpp = instrwrp_cpp.replace("@TI@", str(ti))
        instrwrp_cpp = instrwrp_cpp.replace("@TO@", str(to))
        instrwrp_cpp = instrwrp_cpp.replace("@KO@", str(ko))
        with open(wrapper_output_dir + "/top_instrumentation_wrapper.cpp", "w") as f:
            f.write(instrwrp_cpp)
        # fill out HLS synthesis tcl template
        prjname = "project_instrwrap"
        ipgentcl = ipgentcl_template
        ipgentcl = ipgentcl.replace("$PROJECTNAME$", prjname)
        ipgentcl = ipgentcl.replace("$HWSRCDIR$", wrapper_output_dir)
        ipgentcl = ipgentcl.replace("$FINNHLSLIB$", str(get_deps_path() / "finn-hlslib"))
        ipgentcl = ipgentcl.replace("$ATTENTIONHLSLIB$", str(get_deps_path() / "attention-hlslib"))
        ipgentcl = ipgentcl.replace("$TOPFXN$", "instrumentation_wrapper")
        ipgentcl = ipgentcl.replace("$FPGAPART$", self.fpga_part)
        ipgentcl = ipgentcl.replace("$CLKPERIOD$", str(self.clk_period_ns))
        ipgentcl = ipgentcl.replace("$DEFAULT_DIRECTIVES$", "")
        if self.format == "xo":
            # use Vitis RTL kernel (.xo) output instead of IP-XACT
            ipgentcl = ipgentcl.replace("$EXTRA_DIRECTIVES$", "config_export -format xo")
            ipgentcl = ipgentcl.replace(
                "export_design -format ip_catalog", "export_design -format xo"
            )
        else:
            ipgentcl = ipgentcl.replace("$EXTRA_DIRECTIVES$", "")
        with open(wrapper_output_dir + "/hls_syn.tcl", "w") as f:
            f.write(ipgentcl)
        # build bash script to launch HLS synth and call it
        code_gen_dir = Path(wrapper_output_dir)
        builder = CallHLS(
            tcl_script=code_gen_dir / "hls_syn.tcl",
            code_gen_dir=code_gen_dir,
            ipgen_path=code_gen_dir / prjname,
        )
        builder.build()
        ipgen_path = builder.ipgen_path
        assert ipgen_path.is_dir(), "HLS IPGen failed: %s not found" % (ipgen_path)
        ip_path = ipgen_path / "sol1" / "impl" / "ip"
        assert ip_path.is_dir(), "HLS IPGen failed: %s not found. Check log under %s" % (
            ip_path,
            code_gen_dir,
        )
        if self.format == "xo":
            assert False, "Not implemented"
            # TODO: export for use in VitisBuild or VersalBuild
            # xo_dir = self.output_dir + "/xo"
            # xo_dir = str(os.path.abspath(xo_dir))
            # os.makedirs(xo_dir, exist_ok=True)
            # xo_path = code_gen_dir + "/{}/sol1/impl/export.xo".format(prjname)
            # xo_instr_path = xo_dir + "/instrumentation_wrapper.xo"
            # shutil.copy(xo_path, xo_instr_path)
        else:
            # shutil.move(ip_path, self.output_dir)
            pass

        return (model, False)


class PrepareInstrumentationSim(Transformation):
    """Prepare simulation environment for instrumentation."""

    def __init__(self, fpga_part):
        """Initialize instrumentation simulation preparation."""
        super().__init__()
        self.fpga_part = fpga_part

    def apply(self, model):
        """Prepare scripts for simulating instrumentation IP."""
        # Create directory for simulation of instrumentation IP + FINN IP
        sim_output_dir = make_build_dir(prefix="sim_Instrumentation_")
        model.set_metadata_prop("instrumentation_sim", sim_output_dir)

        # check if instrumentation IP was generated
        instr_ip_dir = model.get_metadata_prop("instrumentation_ipgen")
        if instr_ip_dir is None or (not os.path.isdir(instr_ip_dir)):
            raise Exception(
                "Instrumentation IP not generated, run GenerateInstrumentationIP first."
            )

        # TODO: Support simulation with AXI-lite control interfaces (e.g., for dynamic pipelines)
        # fill in testbench template
        with open(
            os.path.join(os.environ["FINN_CUSTOM_HLS"], "instrumentation_tb.template.sv"),
            "r",
        ) as f:
            testbench_sv = f.read()
        with open(sim_output_dir + "/instrwrap_testbench.sv", "w") as f:
            f.write(testbench_sv)
        # fill in testbench project creator template
        with open(
            os.path.join(os.environ["FINN_CUSTOM_HLS"], "instrumentation_sim.template.tcl"),
            "r",
        ) as f:
            testbench_tcl = f.read()

        # collect ip repo paths for finn accelerator sub cores so Vivado can find them
        ipstitch_path = model.get_metadata_prop("vivado_stitch_proj")
        ip_dirs = ["list"]
        ip_dirs += collect_ip_dirs(model, ipstitch_path)
        ip_dirs += [instr_ip_dir]
        ip_dirs_str = "[%s]" % (" ".join(ip_dirs))
        testbench_tcl = testbench_tcl.replace("@FPGA_PART@", self.fpga_part)
        testbench_tcl = testbench_tcl.replace("@IP_DIRS_STR@", ip_dirs_str)
        with open(sim_output_dir + "/make_instrwrap_sim_proj.tcl", "w") as f:
            f.write(testbench_tcl)

        return (model, False)


class RunInstrumentationSim(Transformation):
    """Run instrumentation simulation to collect performance data."""

    def __init__(self):
        """Initialize instrumentation simulation runner."""
        super().__init__()

    def apply(self, model):
        """Run instrumentation simulation script."""
        sim_output_dir = model.get_metadata_prop("instrumentation_sim")
        if sim_output_dir is None or (not os.path.isdir(sim_output_dir)):
            raise Exception(
                "Instrumentation sim not prepared, run PrepareInstrumentationSim first."
            )

        # Prepare bash script
        bash_script = os.getcwd() + "/report_power.sh"
        with open(bash_script, "w") as script:
            script.write("#!/bin/bash\n")
            script.write("cd %s\n" % (sim_output_dir))
            script.write("vivado -mode batch -source make_instrwrap_sim_proj.tcl\n")

        # Run script
        print("Running Vivado simulation of instrumentation wrapper")
        sub_proc = subprocess.Popen(["bash", bash_script])
        sub_proc.communicate()

        return (model, False)



# Copyright (c) 2020, Xilinx, Inc.
# Copyright (C) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Transformation to build FINN dataflow designs for Alveo using Vitis."""

import json
import os
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import (
    GiveReadableTensorNames,
    GiveUniqueNodeNames,
    RemoveUnusedTensors,
)
from subprocess import CalledProcessError

from finn.builder.build_dataflow_config import FpgaMemoryType, VitisOptStrategy
from finn.transformation.fpgadataflow.create_dataflow_partition import CreateDataflowPartition
from finn.transformation.fpgadataflow.create_stitched_ip import CreateStitchedIP
from finn.transformation.fpgadataflow.floorplan import Floorplan
from finn.transformation.fpgadataflow.hlssynth_ip import HLSSynthIP
from finn.transformation.fpgadataflow.insert_dwc import InsertDWC
from finn.transformation.fpgadataflow.insert_fifo import InsertFIFO
from finn.transformation.fpgadataflow.insert_iodma import InsertIODMA
from finn.transformation.fpgadataflow.prepare_ip import PrepareIP
from finn.transformation.fpgadataflow.specialize_layers import SpecializeLayers
from finn.util.basic import launch_process_helper, make_build_dir
from finn.util.exception import FINNError, FINNUserError
from finn.util.logging import log

from . import templates


def _check_vitis_envvars():
    """Check environment variables for Vitis installation."""
    assert "XILINX_VITIS" in os.environ, "XILINX_VITIS must be set for Vitis"
    assert "PLATFORM_REPO_PATHS" in os.environ, "PLATFORM_REPO_PATHS must be set for Vitis"
    assert (
        "XILINX_XRT" in os.environ
    ), "XILINX_XRT must be set for Vitis, ensure the XRT env is sourced"


class CreateVitisXO(Transformation):
    """Create a Vitis object file from a stitched FINN ip.

    Outcome if successful: sets the vitis_xo attribute in the ONNX
    ModelProto's metadata_props field with the name of the object file as value.
    The object file can be found under the ip subdirectory.
    """

    def __init__(self, ip_name="finn_design"):
        """Initialize CreateVitisXO transformation."""
        super().__init__()
        self.ip_name = ip_name

    def apply(self, model):
        """Apply CreateVitisXO transformation to create Vitis object file."""
        _check_vitis_envvars()
        vivado_proj_dir = model.get_metadata_prop("vivado_stitch_proj")
        stitched_ip_dir = vivado_proj_dir + "/ip"
        interfaces = json.loads(model.get_metadata_prop("vivado_stitch_ifnames"))
        args_string = []
        arg_id = 0
        # NOTE: this assumes the graph is Vitis-compatible: max one axi lite interface
        # developed from instructions in UG1393 (v2019.2) and package_xo documentation
        # package_xo is responsible for generating the kernel xml
        assert len(interfaces["axilite"]) <= 1, "CreateVitisXO supports max 1 AXI lite interface"
        axilite_intf_name = None
        if len(interfaces["axilite"]) == 1:
            axilite_intf_name = interfaces["axilite"][0]
            if len(interfaces["aximm"]) > 0:
                args_string.append(
                    "{addr:1:%s:%s:0x8:0x10:ap_uint&lt;%s>*:0}"
                    % (
                        str(arg_id),
                        interfaces["aximm"][0][0],
                        str(interfaces["aximm"][0][1]),
                    )
                )
                arg_id += 1
                args_string.append(
                    "{numReps:0:%s:%s:0x4:0x1C:uint:0}" % (str(arg_id), axilite_intf_name)
                )
                arg_id += 1
            else:
                args_string.append(
                    "{numReps:0:%s:%s:0x4:0x10:uint:0}" % (str(arg_id), axilite_intf_name)
                )
                arg_id += 1
        for intf in interfaces["s_axis"] + interfaces["m_axis"]:
            stream_width = intf[1]
            stream_name = intf[0]
            args_string.append(
                "{%s:4:%s:%s:0x0:0x0:ap_uint&lt;%s>:0}"
                % (stream_name, str(arg_id), stream_name, str(stream_width))
            )
            arg_id += 1

        # save kernel xml then run package_xo
        xo_name = self.ip_name + ".xo"
        xo_path = vivado_proj_dir + "/" + xo_name
        model.set_metadata_prop("vitis_xo", xo_path)

        # generate the package_xo command in a tcl script
        package_xo_string = "package_xo -force -xo_path %s -kernel_name %s -ip_directory %s" % (
            xo_path,
            self.ip_name,
            stitched_ip_dir,
        )
        for arg in args_string:
            package_xo_string += " -kernel_xml_args " + arg
        with open(vivado_proj_dir + "/gen_xo.tcl", "w") as f:
            f.write(package_xo_string)

        # create a shell script and call Vivado
        package_xo_sh = vivado_proj_dir + "/gen_xo.sh"
        working_dir = os.getcwd()
        with open(package_xo_sh, "w") as f:
            f.write("#!/bin/bash \n")
            f.write("set -e\n")
            f.write("cd {}\n".format(vivado_proj_dir))
            f.write("vivado -mode batch -source gen_xo.tcl\n")
            f.write("cd {}\n".format(working_dir))
        bash_command = ["bash", package_xo_sh]
        try:
            launch_process_helper(bash_command, print_stdout=False)
        except CalledProcessError as e:
            raise FINNUserError(
                f"An error ocurred while generating the XO file for "
                f"{self.ip_name}. Check {vivado_proj_dir} for further "
                f"details."
            ) from e
        if not os.path.isfile(xo_path):
            raise FINNError("Vitis .xo file not created, check logs under %s" % vivado_proj_dir)

        return (model, False)


class VitisLink(Transformation):
    """Create an XCLBIN with Vitis.

    Outcome if successful: sets the bitfile attribute in the ONNX
    ModelProto's metadata_props field with the XCLBIN full path as value.
    """

    def __init__(
        self,
        platform,
        f_mhz=200,
        strategy=VitisOptStrategy.PERFORMANCE,
        enable_debug=False,
        fpga_memory_type="default",
    ):
        """Initialize VitisLink transformation with platform and build settings."""
        super().__init__()
        self.platform = platform
        self.f_mhz = f_mhz
        self.strategy = strategy
        self.enable_debug = enable_debug
        self.fpga_memory_type = fpga_memory_type

    def apply(self, model):
        """Apply VitisLink transformation to create XCLBIN."""
        _check_vitis_envvars()
        # create a config file and empty list of xo files
        config = ["[connectivity]"]
        object_files = []
        idma_idx = 0
        odma_idx = 0
        mem_idx = 0
        instance_names = {}
        for node in model.graph.node:
            assert node.op_type == "StreamingDataflowPartition", "Invalid link graph"
            sdp_node = getCustomOp(node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            kernel_model = ModelWrapper(dataflow_model_filename)
            kernel_xo = kernel_model.get_metadata_prop("vitis_xo")
            object_files.append(kernel_xo)
            # gather info on connectivity
            # assume each node connected to outputs/inputs is DMA:
            # has axis, aximm and axilite
            # everything else is axis-only
            # assume only one connection from each ip to the next
            if len(node.input) == 0:
                producer = None
            else:
                producer = model.find_producer(node.input[0])
            consumer = model.find_consumers(node.output[0])
            # define kernel instances
            # name kernels connected to graph inputs as idmaxx
            # name kernels connected to graph inputs as odmaxx
            # TODO not a good way of checking for external in/out
            # check top-level in/out list instead
            if producer is None:
                instance_names[node.name] = "idma" + str(idma_idx)
                config.append("nk=%s:1:%s" % (node.name, instance_names[node.name]))
                idma_idx += 1
            elif consumer == []:
                instance_names[node.name] = "odma" + str(odma_idx)
                config.append("nk=%s:1:%s" % (node.name, instance_names[node.name]))
                odma_idx += 1
            else:
                instance_names[node.name] = node.name
                config.append("nk=%s:1:%s" % (node.name, instance_names[node.name]))
            sdp_node.set_nodeattr("instance_name", instance_names[node.name])
            # explicitly assign SLRs if the slr attribute is not -1
            node_slr = sdp_node.get_nodeattr("slr")
            if node_slr != -1:
                config.append("slr=%s:SLR%d" % (instance_names[node.name], node_slr))
            # assign memory banks
            if producer is None or consumer is None or consumer == []:
                node_mem_port = sdp_node.get_nodeattr("mem_port")
                if node_mem_port == "":
                    if self.fpga_memory_type == FpgaMemoryType.DEFAULT:
                        # configure good defaults based on board
                        if (
                            "u50" in self.platform
                            or "u280" in self.platform
                            or "u55c" in self.platform
                        ):
                            # Use HBM where available (also U50 does not have DDR)
                            mem_type = "HBM"
                            mem_idx = 0
                        elif "u200" in self.platform:
                            # Use DDR controller in static region of U200
                            mem_type = "DDR"
                            mem_idx = 1
                        elif "u250" in self.platform:
                            # Use DDR controller on the node's SLR if set, otherwise 0
                            mem_type = "DDR"
                            if node_slr == -1:
                                mem_idx = 0
                            else:
                                mem_idx = node_slr
                        else:
                            mem_type = "DDR"
                            mem_idx = 1
                    elif self.fpga_memory_type == FpgaMemoryType.HOST_MEM:
                        mem_type = "HOST"
                        mem_idx = 0
                    else:
                        raise RuntimeError(
                            "Unknown fpga memory type: "
                            + str(self.fpga_memory_type)
                            + ". Aborting!"
                        )
                    node_mem_port = "%s[%d]" % (mem_type, mem_idx)
                config.append("sp=%s.m_axi_gmem0:%s" % (instance_names[node.name], node_mem_port))
            # connect streams
            if producer is not None:
                for i in range(len(node.input)):
                    producer = model.find_producer(node.input[i])
                    if producer is not None:
                        j = list(producer.output).index(node.input[i])
                        config.append(
                            "stream_connect=%s.m_axis_%d:%s.s_axis_%d"
                            % (
                                instance_names[producer.name],
                                j,
                                instance_names[node.name],
                                i,
                            )
                        )

        # create a temporary folder for the project
        link_dir = make_build_dir(prefix="vitis_link_proj_")
        model.set_metadata_prop("vitis_link_proj", link_dir)

        # add Vivado physopt directives if desired
        if self.strategy == VitisOptStrategy.PERFORMANCE_BEST:
            config.append("[vivado]")
            config.append("prop=run.impl_1.STEPS.OPT_DESIGN.ARGS.DIRECTIVE=ExploreWithRemap")
            config.append("prop=run.impl_1.STEPS.PLACE_DESIGN.ARGS.DIRECTIVE=Explore")
            config.append("prop=run.impl_1.STEPS.PHYS_OPT_DESIGN.IS_ENABLED=true")
            config.append("prop=run.impl_1.STEPS.PHYS_OPT_DESIGN.ARGS.DIRECTIVE=Explore")
            config.append("prop=run.impl_1.STEPS.ROUTE_DESIGN.ARGS.DIRECTIVE=Explore")

        config = "\n".join(config) + "\n"
        with open(link_dir + "/config.txt", "w") as f:
            f.write(config)

        # create tcl script to generate resource report in XML format
        gen_rep_xml = templates.vitis_gen_xml_report_tcl_template
        gen_rep_xml = gen_rep_xml.replace("$VITIS_PROJ_PATH$", link_dir)
        with open(link_dir + "/gen_report_xml.tcl", "w") as f:
            f.write(gen_rep_xml)

        debug_commands = []
        if self.enable_debug:
            for inst in list(instance_names.values()):
                debug_commands.append("--dk chipscope:%s" % inst)

        # create a shell script and call Vitis
        script = link_dir + "/run_vitis_link.sh"
        working_dir = os.getcwd()
        with open(script, "w") as f:
            f.write("#!/bin/bash \n")
            f.write("set -e\n")
            f.write("cd {}\n".format(link_dir))
            f.write(
                "v++ -t hw --platform %s --link %s"
                " --kernel_frequency %d --config config.txt --optimize %s"
                " --save-temps -R2 %s\n"
                % (
                    self.platform,
                    " ".join(object_files),
                    self.f_mhz,
                    self.strategy.value,
                    " ".join(debug_commands),
                )
            )
            f.write("cd {}\n".format(working_dir))
        bash_command = ["bash", script]

        try:
            launch_process_helper(bash_command, print_stdout=False)
        except CalledProcessError as e:
            raise FINNUserError(f"Linking failed. Check {link_dir} for further details.") from e
        xclbin = link_dir + "/a.xclbin"
        if not os.path.isfile(xclbin):
            raise FINNError("Vitis .xclbin file not created, check logs under %s" % link_dir)

        # TODO rename xclbin appropriately here?
        model.set_metadata_prop("bitfile", xclbin)

        # run Vivado to gen xml report
        gen_rep_xml_sh = link_dir + "/gen_report_xml.sh"
        working_dir = os.getcwd()
        with open(gen_rep_xml_sh, "w") as f:
            f.write("#!/bin/bash \n")
            f.write("cd {}\n".format(link_dir))
            f.write("set -e\n")
            f.write("vivado -mode batch -source %s\n" % (link_dir + "/gen_report_xml.tcl"))
            f.write("cd {}\n".format(working_dir))
        bash_command = ["bash", gen_rep_xml_sh]
        try:
            launch_process_helper(bash_command, print_stdout=False)
        except CalledProcessError:
            log.error(f"Creation of XML reports failed. Check {link_dir} for details. Continuing..")
        # filename for the synth utilization report
        synth_report_filename = link_dir + "/synth_report.xml"
        model.set_metadata_prop("vivado_synth_rpt", synth_report_filename)
        return (model, False)


class VitisBuild(Transformation):
    """Best-effort attempt at building the accelerator with Vitis.
    It assumes the model has only fpgadataflow nodes

    :parameter fpga_part: string identifying the target FPGA
    :parameter period_ns: target clock period
    :parameter platform: target Alveo platform, one of ["U50", "U200", "U250", "U280"]
    :parameter strategy: Vitis optimization strategy
    :parameter enable_debug: add Chipscope to all AXI interfaces
    :parameter floorplan_file: path to a JSON containing a dictionary with
        SLR assignments for each node in the ONNX graph.
        Must be parse-able by the ApplyConfig transform.
    :parameter enable_link: enable linking kernels (.xo files),
        otherwise just synthesize them independently.
    :parameter fpga_memory_type: Specify whether Host or FPGA memory such as DDR/HBM should be used
    """

    def __init__(
        self,
        fpga_part,
        period_ns,
        platform,
        strategy=VitisOptStrategy.PERFORMANCE,
        enable_debug=False,
        floorplan_file=None,
        enable_link=True,
        partition_model_dir=None,
        fpga_memory_type=FpgaMemoryType.DEFAULT,
    ):
        """Initialize VitisBuild transformation with FPGA and build settings."""
        super().__init__()
        self.fpga_part = fpga_part
        self.period_ns = period_ns
        self.platform = platform
        self.strategy = strategy
        self.enable_debug = enable_debug
        self.floorplan_file = floorplan_file
        self.enable_link = enable_link
        self.partition_model_dir = partition_model_dir
        self.fpga_memory_type = fpga_memory_type

    def apply(self, model):
        """Apply VitisBuild transformation to create complete Vitis accelerator."""
        _check_vitis_envvars()
        # prepare at global level, then break up into kernels
        prep_transforms = [InsertIODMA(512), InsertDWC(), SpecializeLayers(self.fpga_part)]
        for trn in prep_transforms:
            model = model.transform(trn)
            model = model.transform(GiveUniqueNodeNames())
            model = model.transform(GiveReadableTensorNames())

        model = model.transform(Floorplan(floorplan=self.floorplan_file))

        model = model.transform(
            CreateDataflowPartition(partition_model_dir=self.partition_model_dir)
        )
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())

        # Build each kernel individually
        sdp_nodes = model.get_nodes_by_op_type("StreamingDataflowPartition")
        for sdp_node in sdp_nodes:
            prefix = sdp_node.name + "_"
            sdp_node = getCustomOp(sdp_node)
            dataflow_model_filename = sdp_node.get_nodeattr("model")
            kernel_model = ModelWrapper(dataflow_model_filename)
            kernel_model = kernel_model.transform(InsertFIFO())
            kernel_model = kernel_model.transform(SpecializeLayers(self.fpga_part))
            kernel_model = kernel_model.transform(RemoveUnusedTensors())
            kernel_model = kernel_model.transform(GiveUniqueNodeNames(prefix))
            kernel_model.save(dataflow_model_filename)
            kernel_model = kernel_model.transform(PrepareIP(self.fpga_part, self.period_ns))
            kernel_model = kernel_model.transform(HLSSynthIP())
            kernel_model = kernel_model.transform(
                CreateStitchedIP(self.fpga_part, self.period_ns, sdp_node.onnx_node.name, True)
            )
            kernel_model = kernel_model.transform(CreateVitisXO(sdp_node.onnx_node.name))
            kernel_model.set_metadata_prop("platform", "alveo")
            kernel_model.save(dataflow_model_filename)
        # Assemble design from kernels
        if self.enable_link:
            model = model.transform(
                VitisLink(
                    self.platform,
                    round(1000 / self.period_ns),
                    strategy=self.strategy,
                    enable_debug=self.enable_debug,
                    fpga_memory_type=self.fpga_memory_type,
                )
            )
        # set platform attribute for correct remote execution
        model.set_metadata_prop("platform", "alveo")

        return (model, False)



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto
from onnx import helper as oh
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_by_name


class InsertTLastMarker(Transformation):
    """Ensure that the graph is started/terminated with a TLastMarker_hls node, inserting
    one if necessary.
    Use constructor args to determine type of TLastMarker to be inserted.
    More information available on the TLastMarker documentation.
    """

    def __init__(self, both=False, external=True, dynamic=True):
        super().__init__()
        self.dyniters = dynamic
        self.external = external
        self.both = both

    def apply(self, model):
        # TODO only makes sense for a pure fpgadataflow graph -- check!
        graph_out_name = model.graph.output[0].name
        final_node = model.find_producer(graph_out_name)
        graph_modified = False
        if final_node.op_type != "TLastMarker_hls" and not (
            final_node.op_type == "IODMA_hls"
            and get_by_name(final_node.attribute, "direction").s.decode("UTF-8") == "out"
        ):
            custom_op = getCustomOp(final_node)
            num_iters = int(custom_op.get_number_output_values())
            stream_width = int(custom_op.get_outstream_width())
            out_shape = model.get_tensor_shape(graph_out_name)
            out_dtype = model.get_tensor_datatype(graph_out_name)
            elem_width = out_dtype.bitwidth()
            # make new buffer
            final_node_out = oh.make_tensor_value_info(
                model.make_new_valueinfo_name(), TensorProto.FLOAT, out_shape
            )
            model.graph.value_info.append(final_node_out)
            model.set_tensor_datatype(final_node_out.name, out_dtype)
            # reroute final node output to final_node_out_name
            final_node.output[0] = final_node_out.name
            tlast_node = oh.make_node(
                "TLastMarker_hls",
                [final_node_out.name],
                [graph_out_name],
                NumIters=num_iters,
                StreamWidth=stream_width,
                ElemWidth=elem_width,
                DynIters=(1 if self.dyniters else 0),
                Direction="out",
                Protocol=("external" if self.external else "internal"),
                domain="finn.custom_op.fpgadataflow.hls",
                backend="fpgadataflow",
            )
            model.graph.node.append(tlast_node)
            graph_modified = True
        # if both is True, also insert marker on input
        if self.both:
            # detect and parse graph inputs
            insert_idx = 0
            graph_in_names = [x.name for x in model.graph.input]
            for graph_in_name in graph_in_names:
                first_node = model.find_consumers(graph_in_name)
                # skip if no consumers (this may be the case for unused initializers)
                # TODO: fix this with a cleanup transform
                if first_node == []:
                    continue
                assert len(first_node) == 1, "Input fans out to multiple nodes"
                first_node = first_node[0]
                # several scenarios exclude the node:
                # 1. node is a FC layer with internal weights, in which case
                #    the input is in the list of graph inputs because it has an
                #    initializer (TODO: fix this with a clean-up transform)
                if (
                    first_node.op_type.startswith("MVAU")
                    and get_by_name(first_node.attribute, "mem_mode").s.decode("UTF-8")
                    != "external"
                ):
                    continue
                # 2. node is either a TLastMarker or an input IODMA
                if first_node.op_type != "TLastMarker_hls" and not (
                    first_node.op_type == "IODMA_hls"
                    and get_by_name(first_node.attribute, "direction").s.decode("UTF-8") == "in"
                ):
                    custom_op = getCustomOp(first_node)
                    num_iters = np.prod(custom_op.get_folded_input_shape()[1:-1])
                    inp_idx = list(first_node.input).index(graph_in_name)
                    if inp_idx > 0:
                        if first_node.op_type.startswith("MVAU") and inp_idx == 1:
                            stream_width = int(custom_op.get_instream_width(1))
                        elif first_node.op_type.startswith("AddStreams") and inp_idx == 1:
                            stream_width = int(custom_op.get_instream_width())
                        else:
                            raise Exception("No method to determine stream width")
                    else:
                        stream_width = int(custom_op.get_instream_width())
                    in_shape = model.get_tensor_shape(graph_in_name)
                    in_dtype = model.get_tensor_datatype(graph_in_name)
                    elem_width = in_dtype.bitwidth()
                    # make new buffer
                    first_node_in = oh.make_tensor_value_info(
                        model.make_new_valueinfo_name(), TensorProto.FLOAT, in_shape
                    )
                    model.graph.value_info.append(first_node_in)
                    model.set_tensor_datatype(first_node_in.name, in_dtype)
                    ini = model.get_initializer(graph_in_name)
                    # copy initializer if it exists
                    if ini is not None:
                        model.set_initializer(first_node_in.name, ini)
                    # reroute final node output to first_node_in_name
                    first_node.input[inp_idx] = first_node_in.name
                    tlast_node = oh.make_node(
                        "TLastMarker_hls",
                        [graph_in_name],
                        [first_node_in.name],
                        NumIters=num_iters,
                        StreamWidth=stream_width,
                        ElemWidth=elem_width,
                        DynIters=(1 if self.dyniters else 0),
                        Direction="in",
                        Protocol=("external" if self.external else "internal"),
                        domain="finn.custom_op.fpgadataflow.hls",
                        backend="fpgadataflow",
                    )
                    model.graph.node.insert(insert_idx, tlast_node)
                    graph_modified = True
                    insert_idx += 1
        return (model, graph_modified)



import argparse
import json
import time
from pynq import PL, Overlay
from pynq.pl_server.device import Device
from pynq.ps import Clocks

# Instrumentation wrapper register map #
# ap_uint<32>  cfg,   	// [0] - 0:hold, 1:lfsr; [31:16] - LFSR seed
# ap_uint<32> &status,	// [0] - timestamp overflow; [1] - timestamp underflow
# ap_uint<32> &latency,
# ap_uint<32> &interval,
# ap_uint<32> &checksum,
# ap_uint<32> &min_latency


class FINNInstrumentationOverlay(Overlay):
    def __init__(
        self,
        bitfile_name,
        platform="zynq",
        fclk_mhz=100.0,
        device=None,
        download=True,
        seed=1,
    ):
        super().__init__(bitfile_name, download=download, device=device)

        self.platform = platform
        self.fclk_mhz = fclk_mhz
        self.seed = seed

        # configure clock (for ZYNQ platforms)
        if self.platform == "zynq":
            if self.fclk_mhz > 0:
                Clocks.fclk0_mhz = self.fclk_mhz
                self.fclk_mhz_actual = Clocks.fclk0_mhz

    def instrumentation_read(self, name):
        return self.instrumentation_wrap_0.read(
            offset=self.ip_dict["instrumentation_wrap_0"]["registers"][name]["address_offset"]
        )

    def instrumentation_write(self, name, value):
        return self.instrumentation_wrap_0.write(
            offset=self.ip_dict["instrumentation_wrap_0"]["registers"][name]["address_offset"],
            value=value,
        )

    def reset_accelerator(self):
        self.axi_gpio_0.write(
            offset=self.ip_dict["axi_gpio_0"]["registers"]["GPIO_DATA"]["address_offset"], value=0
        )

    def start_accelerator(self):
        lfsr_seed = (self.seed << 16) & 0xFFFF0000  # upper 16 bits
        self.instrumentation_write("cfg", lfsr_seed + 1)  # start operation

    def observe_instrumentation(self, debug_print=True):
        status_reg = self.instrumentation_read("status")
        chksum_reg = self.instrumentation_read("checksum")
        min_latency = self.instrumentation_read("min_latency")
        latency = self.instrumentation_read("latency")
        interval = self.instrumentation_read("interval")

        frame = (chksum_reg >> 24) & 0x000000FF
        checksum = chksum_reg & 0x00FFFFFF
        overflow_err = (status_reg & 0x00000001) != 0
        underflow_err = (status_reg & 0x00000002) != 0

        if debug_print:
            print("---INSTRUMENTATION_REPORT---")
            if overflow_err or underflow_err:
                print("Status ERROR")
                print("Overflow error: %s" % overflow_err)
                print("Underflow error: %s" % underflow_err)
            else:
                print("Status OK")
            print("Frame number (8-bit): %d" % frame)
            print("Checksum: 0x%06x" % checksum)
            print("Min Latency (cycles): %d" % min_latency)
            print("Latency (cycles): %d" % latency)
            print("Interval (cycles): %d" % interval)
            print("----------------------------")

        return (overflow_err, underflow_err, frame, checksum, min_latency, latency, interval)


def run_idle(*args, **kwargs):
    # Program FPGA without running accelerator. Only used in the context of power measurement
    runtime = kwargs["runtime"]
    frequency = kwargs["frequency"]
    seed = kwargs["seed"]
    bitfile = kwargs["bitfile"]
    settingsfile = kwargs["settingsfile"]
    devID = kwargs["device"]

    device = Device.devices[devID]

    # overwrite frequency if specified in settings file
    if settingsfile != "":
        with open(settingsfile, "r") as f:
            settings = json.load(f)
            if "fclk_mhz" in settings:
                frequency = settings["fclk_mhz"]

    print("Programming FPGA..")
    PL.reset()  # reset PYNQ cache
    FINNInstrumentationOverlay(bitfile_name=bitfile, device=device, fclk_mhz=frequency, seed=seed)

    print("Running idle for %d seconds.." % runtime)
    time.sleep(runtime)
    print("Done.")


def main(*args, **kwargs):
    runtime = kwargs["runtime"]
    frequency = kwargs["frequency"]
    seed = kwargs["seed"]
    bitfile = kwargs["bitfile"]
    reportfile = kwargs["reportfile"]
    settingsfile = kwargs["settingsfile"]
    devID = kwargs["device"]

    device = Device.devices[devID]

    # overwrite frequency if specified in settings file
    if settingsfile != "":
        with open(settingsfile, "r") as f:
            settings = json.load(f)
            if "fclk_mhz" in settings:
                frequency = settings["fclk_mhz"]

    # instantiate FINN accelerator driver and pass batchsize and bitfile
    print("Programming FPGA..")
    PL.reset()  # reset PYNQ cache
    accel = FINNInstrumentationOverlay(
        bitfile_name=bitfile, device=device, fclk_mhz=frequency, seed=seed
    )

    # start accelerator
    print("Running accelerator for %d seconds.." % runtime)
    accel.start_accelerator()

    # let it run for specified runtime
    time.sleep(runtime)

    # read measurement from instrumentation
    (
        overflow_err,
        underflow_err,
        frame,
        checksum,
        min_latency,
        latency,
        interval,
    ) = accel.observe_instrumentation()

    # write report to file
    report = {
        "error": overflow_err or underflow_err or interval == 0,
        "checksum": checksum,
        "min_latency_cycles": min_latency,
        "latency_cycles": latency,
        "interval_cycles": interval,
        "frequency_mhz": round(accel.fclk_mhz_actual),
        "min_latency_ms": round(min_latency * (1 / (accel.fclk_mhz_actual * 1e6)) * 1e3, 6),
        "latency_ms": round(latency * (1 / (accel.fclk_mhz_actual * 1e6)) * 1e3, 6),
        "throughput_fps": round(1 / (interval * (1 / (accel.fclk_mhz_actual * 1e6)))),
        "min_pipeline_depth": round(min_latency / interval, 2),
        "pipeline_depth": round(latency / interval, 2),
    }
    with open(reportfile, "w") as f:
        json.dump(report, f, indent=2)

    print("Done.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Profile FINN-generated accelerator using instrumentation wrapper"
    )
    parser.add_argument("--runtime", help="Runtime in seconds", type=int, default=10)
    parser.add_argument(
        "--frequency", help="FPGA clock frequency in MHz", type=float, default=100.0
    )
    parser.add_argument("--seed", help="LFSR seed for input data generation", type=int, default=1)
    parser.add_argument("--device", help="FPGA device to be used", type=int, default=0)
    parser.add_argument("--bitfile", help="Name of bitfile", default="finn-accel.bit")
    parser.add_argument(
        "--reportfile",
        help="Name of output .json report file",
        type=str,
        default="measured_performance.json",
    )
    parser.add_argument(
        "--settingsfile", help="Name of optional input .json settings file", type=str, default=""
    )
    args = parser.parse_args()
    main([], vars(args))



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of FINN nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import qonnx.custom_op.registry as registry
from qonnx.util.basic import is_finn_op


def aggregate_dict_keys(res_dict):
    total_dict = {}
    for layer in res_dict:
        layer_res_dict = res_dict[layer]
        for r_type in layer_res_dict.keys():
            if "efficiency" in r_type:
                continue
            r_amount = layer_res_dict[r_type]
            r_amount = float(r_amount)
            if r_type in total_dict.keys():
                total_dict[r_type] += r_amount
            else:
                total_dict[r_type] = r_amount
    return total_dict


def op_and_param_counts(model):
    """Return per-node and aggregate op counts per inference."""

    ret_dict = {}
    for node in model.graph.node:
        if is_finn_op(node.domain):
            inst = registry.getCustomOp(node)
            if hasattr(inst, "get_op_and_param_counts"):
                node_op_and_param_counts = inst.get_op_and_param_counts()
                ret_dict[node.name] = node_op_and_param_counts
    ret_dict["total"] = aggregate_dict_keys(ret_dict)
    return ret_dict



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
FINN dataflow build system.

This module provides the main build infrastructure for converting ONNX models
to FINN dataflow accelerators. It handles step resolution, logging, error handling,
and the complete build pipeline from ONNX input to hardware accelerator output.
"""

import datetime
import importlib
import json
import logging
import os
from pathlib import Path
from typing import Callable

import pdb  # isort: split
import sys
import time
from qonnx.core.modelwrapper import ModelWrapper
from rich import print as rprint
from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import Traceback

from finn.builder.build_dataflow_config import DataflowBuildConfig, default_build_dataflow_steps
from finn.builder.build_dataflow_steps import build_dataflow_step_lookup
from finn.util.exception import (
    FINNConfigurationError,
    FINNDataflowError,
    FINNError,
    FINNUserError,
    snapshot_on_exception,
)


def get_logfile_path(cfg: DataflowBuildConfig) -> str:
    """Return the path to the logfile in the build dir."""
    return str(Path(cfg.output_dir) / "build_dataflow.log")


# adapted from https://stackoverflow.com/a/39215961
class PrintLogger(object):
    """Custom stream handler that writes to both console and log file with timestamps."""

    def __init__(self, logger, level, originalstream):
        """Initialize the print logger with logger, level, and original stream."""
        self.logger = logger
        self.level = level
        self.console = originalstream
        self.linebuf = ""

    def write(self, buf):
        """Write buffer content to both logger and console with timestamp."""
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())
            timestamp = datetime.datetime.fromtimestamp(time.time()).strftime("%Y-%m-%d %H:%M:%S")
            self.console.write(f"[{timestamp}] " + line + "\n")

    def flush(self):
        """Flush the console stream."""
        self.console.flush()


def resolve_build_steps(cfg: DataflowBuildConfig, partial: bool = True) -> list[Callable]:
    """
    Resolve build step names to callable functions.

    Converts string step names to callable functions by looking up in the step registry,
    importing from modules, or using direct callable objects. Supports partial execution
    between start_step and stop_step if specified in config.

    Args:
        cfg: Build configuration containing step definitions
        partial: If True, respect start_step/stop_step boundaries

    Returns:
        List of callable step functions ready for execution
    """
    steps = cfg.steps
    if steps is None:
        steps = default_build_dataflow_steps
    steps_as_fxns = []
    for transform_step in steps:
        if type(transform_step) is str:
            # lookup step function from step name
            if transform_step in build_dataflow_step_lookup.keys():
                steps_as_fxns.append(build_dataflow_step_lookup[transform_step])
            else:
                if "." not in transform_step:
                    if transform_step not in globals().keys():
                        msg = (
                            f"Step {transform_step} is not a default step, not in globals() "
                            "and not an importable name!"
                        )
                        raise FINNConfigurationError(msg)
                    else:  # noqa
                        fxn_step = globals()[transform_step]
                        if not callable(fxn_step):
                            msg = (
                                f"Step {transform_step} was resolved in globals(), but is "
                                "not callable object. If the name was already in use, consider "
                                "moving your custom step into it's own module and importing it "
                                "via yourmodule.yourstep!"
                            )
                            raise FINNConfigurationError(msg)
                        steps_as_fxns.append(fxn_step)
                        continue
                else:
                    split_step = transform_step.split(".")
                    module_path, fxn_step_name = split_step[:-1], split_step[-1]
                    try:
                        imported_module = importlib.import_module(".".join(module_path))
                        fxn_step = getattr(imported_module, fxn_step_name)
                        if callable(fxn_step):
                            steps_as_fxns.append(fxn_step)
                            continue
                        else:  # noqa
                            msg = (
                                f"Could import custom step module, but final name is not a "
                                f"callable object. Path was {transform_step}"
                            )
                            raise FINNConfigurationError(msg)
                    except ModuleNotFoundError as mnf:
                        msg = (
                            f"Could not resolve build step: {transform_step}. "
                            "The given step is neither importable nor a default step."
                        )
                        raise FINNConfigurationError(msg) from mnf
        elif callable(transform_step):
            # treat step as function to be called as-is
            steps_as_fxns.append(transform_step)
        else:
            raise FINNConfigurationError("Could not resolve build step: " + str(transform_step))
    if partial:
        step_names = list(map(lambda x: x.__name__, steps_as_fxns))
        if cfg.start_step is None:
            start_ind = 0
        else:
            start_ind = step_names.index(cfg.start_step)
        if cfg.stop_step is None:
            stop_ind = len(step_names) - 1
        else:
            stop_ind = step_names.index(cfg.stop_step)
        steps_as_fxns = steps_as_fxns[start_ind : (stop_ind + 1)]

    # Add the exception snapshot decorator if needed
    return [
        snapshot_on_exception(
            snapshot_finn=False, snapshot_config=True, snapshot_buildlog=True, build_dir_prefix=None
        )(transform_step)
        if (
            cfg.enable_exception_snapshots
            and "snapshot_on_exception_enabled" not in dir(transform_step)
        )
        else transform_step
        for transform_step in steps_as_fxns
    ]


def resolve_step_filename(step_name: str, cfg: DataflowBuildConfig, step_delta: int = 0):
    """
    Resolve the intermediate model filename for a given build step.

    Args:
        step_name: Name of the build step
        cfg: Build configuration
        step_delta: Offset from the step (0=current, -1=previous, +1=next)

    Returns:
        Path to the intermediate model file for the specified step
    """
    step_names = list(map(lambda x: x.__name__, resolve_build_steps(cfg, partial=False)))
    if step_name not in step_names:
        raise FINNConfigurationError(
            f"Cannot restart from step {step_name}.Step {step_name} for restarting not found."
        )
    step_no = step_names.index(step_name) + step_delta
    if step_no < 0 or step_no >= len(step_names):
        raise FINNDataflowError("Invalid step+delta combination")
    filename = cfg.output_dir + "/intermediate_models/"
    filename += "%s.onnx" % (step_names[step_no])
    if not Path(filename).exists():
        raise FINNConfigurationError(
            f"Expected model file at {filename} to start from step "
            f"{step_name}, but could not find it!"
        )
    return filename


def setup_logging(cfg: DataflowBuildConfig):
    """
    Configure logging for the build process.

    Sets up file logging, console mirroring, and rich console handlers
    based on the build configuration settings.

    Args:
        cfg: Build configuration with logging settings

    Returns:
        Configured logger instance
    """
    # Set up global logger, the force=True has the following effects:
    # - If multiple build are run in a row, the log file will be re-created for each,
    #   which is needed if the file was deleted/moved or the output dir changed
    # - In a PyTest session, this logger will replace the PyTest log handlers, so logs
    #   (+ captured warnings!) will end up in the log file instead of being collected by PyTest
    logpath = get_logfile_path(cfg)
    if cfg.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format="[%(asctime)s]%(levelname)s: %(pathname)s:%(lineno)d: %(message)s",
            filename=logpath,
            filemode="w",
            force=True,
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format="[%(asctime)s]%(levelname)s: %(message)s",
            filename=logpath,
            filemode="w",
            force=True,
        )

    # Capture all warnings.warn calls of qonnx, ...
    logging.captureWarnings(True)

    # Mirror stdout and stderr to log
    log = logging.getLogger("build_dataflow")
    # Explicitly set the logger level based on cfg.verbose
    if cfg.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)
    if not isinstance(sys.stdout, PrintLogger):
        # Prevent rediricting stdout/sterr multiple times
        sys.stdout = PrintLogger(log, logging.INFO, sys.stdout)
        sys.stderr = PrintLogger(log, logging.ERROR, sys.stderr)
    console = Console(file=sys.stdout.console)

    # Mirror a configurable log level to console (default = ERROR)
    if cfg.console_log_level != "NONE":
        consoleHandler = RichHandler(
            show_time=True, log_time_format="[%Y-%m-%d %H:%M:%S]", show_path=False, console=console
        )
        if cfg.console_log_level == "DEBUG":
            consoleHandler.setLevel(logging.DEBUG)
        elif cfg.console_log_level == "INFO":
            consoleHandler.setLevel(logging.INFO)
        elif cfg.console_log_level == "WARNING":
            consoleHandler.setLevel(logging.WARNING)
        elif cfg.console_log_level == "ERROR":
            consoleHandler.setLevel(logging.ERROR)
        elif cfg.console_log_level == "CRITICAL":
            consoleHandler.setLevel(logging.CRITICAL)
        logging.getLogger().addHandler(consoleHandler)

    return log


def exit_buildflow(cfg: DataflowBuildConfig, time_per_step: dict = None, exit_code: int = 0):
    """
    Handle build completion and generate metadata files.

    Creates metadata_builder.json and time_per_step.json files with build results.

    Args:
        cfg: Build configuration
        time_per_step: Dictionary of step execution times
        exit_code: Build exit code (0=success, non-zero=failure)

    Returns:
        The provided exit code
    """
    if exit_code:
        print("Build failed")
        status = "failed"
    else:
        print("Build completed successfully")
        status = "ok"

    # Generate metadata_builder.json
    metadata = {
        "status": status,
        "tool_version": os.path.basename(os.environ.get("XILINX_VIVADO")),
    }
    with open(os.path.join(cfg.output_dir, "report/metadata_builder.json"), "w") as f:
        json.dump(metadata, f, indent=2)

    # Generate time_per_step.json
    if time_per_step is not None:
        time_per_step["total_build_time"] = sum(time_per_step.values())
        with open(os.path.join(cfg.output_dir, "report/time_per_step.json"), "w") as f:
            json.dump(time_per_step, f, indent=2)

    return exit_code


def build_dataflow_cfg(model_filename, cfg: DataflowBuildConfig):
    """
    Build a dataflow accelerator using the given configuration.

    Main entry point for building FINN dataflow accelerators. Handles step execution,
    logging, error handling, and intermediate model saving.

    Args:
        model_filename: ONNX model filename to build
        cfg: Build configuration specifying steps and options

    Returns:
        Exit code (0=success, non-zero=failure)
    """
    # Create the output (report) dir if it doesn't exist
    os.makedirs(os.path.join(cfg.output_dir, "report"), exist_ok=True)

    log = setup_logging(cfg)
    logfile = get_logfile_path(cfg)
    print(f"Intermediate outputs will be generated in {os.environ['FINN_BUILD_DIR']}")
    print(f"Final outputs will be generated in {cfg.output_dir}")
    print(f"Build log is at {logfile}")

    # Setup done, start build flow
    try:
        # If start_step is specified, override the input model
        if cfg.start_step is None:
            print(f"Building dataflow accelerator from {model_filename}")
            model = ModelWrapper(model_filename)
        else:
            if model_filename != "":
                log.warning(
                    "When using a start-step, FINN automatically searches "
                    "for the correct model to use from previous runs, overwriting your "
                    "passed model file (but still using it's path for the location of the "
                    "temporary file directory, etc.). This behaviour might change "
                    "in future versions!"
                )
            intermediate_model_filename = resolve_step_filename(cfg.start_step, cfg, -1)
            out = (
                f"Building dataflow accelerator from intermediate"
                f" checkpoint {intermediate_model_filename}"
            )
            print(out)
            model = ModelWrapper(intermediate_model_filename)
        assert type(model) is ModelWrapper

        # Start processing
        step_num = 1
        time_per_step = dict()
        build_dataflow_steps = resolve_build_steps(cfg)

        for transform_step in build_dataflow_steps:
            step_name = transform_step.__name__
            print(f"Running step: {step_name} [{step_num}/{len(build_dataflow_steps)}]")

            # Run the step
            step_start = time.time()
            model = transform_step(model, cfg)
            step_end = time.time()
            time_per_step[step_name] = round(step_end - step_start)
            chkpt_name = f"{step_name}.onnx"
            if cfg.save_intermediate_models:
                intermediate_model_dir = os.path.join(cfg.output_dir, "intermediate_models")
                if not os.path.exists(intermediate_model_dir):
                    os.makedirs(intermediate_model_dir)
                model.save(os.path.join(intermediate_model_dir, chkpt_name))
            step_num += 1
    except KeyboardInterrupt:
        print("KeyboardInterrupt detected. Aborting...")
        return exit_buildflow(cfg, time_per_step, -1)
    except (Exception, FINNError) as e:
        # Re-raise exception if we are in a PyTest session so we don't miss it
        if "PYTEST_CURRENT_TEST" in os.environ:
            raise

        if issubclass(type(e), FINNUserError):
            # Handle FINN USER ERROR
            log.error(f"FINN ERROR: {e}")
        else:
            # Handle remaining errors (= FINN INTERNAL COMPILER ERROR)
            log.error(f"FINN INTERNAL COMPILER ERROR: {e}")

        # Print traceback for interal errors or if in debug mode
        if not issubclass(type(e), FINNUserError) or log.level == logging.DEBUG:
            # Restoring stdout and stderr
            if type(sys.stdout) is PrintLogger:
                sys.stdout = sys.stdout.console
            if type(sys.stderr) is PrintLogger:
                sys.stderr = sys.stderr.console

            # Print traceback both to console and logfile
            rprint(Traceback(show_locals=False))
            with Path(get_logfile_path(cfg)).open("a") as f:
                rprint(Traceback(show_locals=False), file=f)

            # Start postmortem debug if configured
            if cfg.enable_build_pdb_debug:
                pdb.post_mortem(e.__traceback__)

        return exit_buildflow(cfg, time_per_step, -1)
    return exit_buildflow(cfg, time_per_step, 0)


def build_dataflow_directory(path_to_cfg_dir: str):
    """
    Build a dataflow accelerator from a directory containing model and config.

    Args:
        path_to_cfg_dir: Directory containing model.onnx and dataflow_build_config.json

    Returns:
        Exit code from build process

    The directory must contain:
    - model.onnx: ONNX model to be converted to dataflow accelerator
    - dataflow_build_config.json: JSON file with build configuration
    """
    # get absolute path
    path_to_cfg_dir = os.path.abspath(path_to_cfg_dir)
    assert os.path.isdir(path_to_cfg_dir), "Directory not found: " + path_to_cfg_dir
    onnx_filename = path_to_cfg_dir + "/model.onnx"
    json_filename = path_to_cfg_dir + "/dataflow_build_config.json"
    assert os.path.isfile(onnx_filename), "ONNX not found: " + onnx_filename
    assert os.path.isfile(json_filename), "Build config not found: " + json_filename
    with open(json_filename, "r") as f:
        json_str = f.read()
    build_cfg = DataflowBuildConfig.from_json(json_str)
    old_wd = os.getcwd()
    # change into build dir to resolve relative paths
    os.chdir(path_to_cfg_dir)
    ret = build_dataflow_cfg(onnx_filename, build_cfg)
    os.chdir(old_wd)
    return ret



import os
import shlex
import subprocess
import sys
from pathlib import Path

from finn.interface import IS_POSIX
from finn.interface.interface_utils import status


def run_test(variant: str, num_workers: str) -> None:
    """Run a given test variant with the given number of workers"""
    original_dir = Path.cwd()

    # TODO: Make this optional
    if "CI_PROJECT_DIR" in os.environ.keys():
        ci_project_dir = os.environ["CI_PROJECT_DIR"]
    else:
        ci_project_dir = os.environ["FINN_BUILD_DIR"]
    status(f"Putting test reports into {ci_project_dir}")

    os.chdir(os.environ["FINN_TESTS"])
    match variant:
        case "quick":
            subprocess.run(
                shlex.split(
                    f"{sys.executable} -m pytest -v -m 'not "
                    f"(vivado or slow or vitis or board or notebooks or bnn_pynq or end2end)' "
                    f"--dist=loadfile -n {num_workers}",
                    posix=IS_POSIX,
                )
            )
        case "quicktest_ci":
            subprocess.run(
                shlex.split(
                    f"{sys.executable} -m pytest -v -m 'not "
                    f"(vivado or slow or vitis or board or notebooks or bnn_pynq or end2end)' "
                    f"--junitxml={ci_project_dir}/reports/quick.xml "
                    f"--html={ci_project_dir}/reports/quick.html "
                    f"--reruns 1 --dist worksteal -n {num_workers}",
                    posix=IS_POSIX,
                )
            )
        case "full_ci":
            test_1_process = subprocess.Popen(
                shlex.split(
                    (
                        f"{sys.executable} -m pytest -v -m 'not "
                        f"(end2end or sanity_bnn or notebooks)' "
                        f"--junitxml={ci_project_dir}/reports/main.xml "
                        f"--html={ci_project_dir}/reports/main.html "
                        f"--reruns 1 --dist worksteal -n {num_workers}"
                    ),
                    posix=IS_POSIX,
                )
            )
            test_2_process = subprocess.Popen(
                shlex.split(
                    (
                        f"{sys.executable} -m pytest -v -m 'end2end or sanity_bnn or notebooks' "
                        f"--junitxml={ci_project_dir}/reports/end2end.xml "
                        f"--html={ci_project_dir}/reports/end2end.html "
                        f"--reruns 1 --dist loadgroup -n {num_workers}"
                    ),
                    posix=IS_POSIX,
                )
            )
            test_1_process.communicate()
            test_1_returncode = test_1_process.returncode
            test_2_process.communicate()
            test_2_returncode = test_2_process.returncode

            subprocess.run(
                shlex.split(
                    (
                        f"{sys.executable} -m pytest_html_merger -i {ci_project_dir}/reports/ "
                        f"-o {ci_project_dir}/reports/full_test_suite.html"
                    ),
                    posix=IS_POSIX,
                )
            )

            if test_1_returncode or test_2_returncode:
                sys.exit(1)

        case _:
            subprocess.run(
                shlex.split(f"{sys.executable} -m pytest -k '{variant}'", posix=IS_POSIX)
            )
    os.chdir(original_dir)



"""
Transformer-specific build steps for FINN.

ADAPTED FROM Christoph's radioml-transformer repository, specifically these files:
build_steps.py
custom/apply_config.py

This module contains custom build steps and transformations specifically designed
for transformer neural network architectures, including attention mechanisms,
elementwise operations, and various hardware conversion steps.
"""

import json

# YAML for loading experiment configurations
import yaml

# QONNX quantization data types
from qonnx.core.datatype import DataType

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# Converts ONNX graph nodes to QONNX custom-ops if possible
from qonnx.custom_op.registry import getCustomOp

# Converts BatchNorm operation to affine transformation
from qonnx.transformation.batchnorm_to_affine import BatchNormToAffine

# Transformation for exhaustively composing transformations
from qonnx.transformation.composed import ComposedTransformation

# If we have a convolution with a bias tensors input, QONNX and later FINN
# expect the bias to be expressed as a standalone Add node following the Conv
# node.
from qonnx.transformation.extract_conv_bias import ExtractBiasFromConv

# Collapses chains of constants into a single constant operation or even
# initializer tensors.
from qonnx.transformation.fold_constants import FoldConstants

# Converts Gemm operation to MatMul with extracted standalone bias op
from qonnx.transformation.gemm_to_matmul import GemmToMatMul

# QONNX graph transformations for renaming and cleaning up
from qonnx.transformation.general import (
    GiveReadableTensorNames,
    GiveUniqueNodeNames,
    GiveUniqueParameterTensors,
    RemoveStaticGraphInputs,
    RemoveUnusedTensors,
    Transformation,
)

# QONNX graph transformations for annotating the graph with datatype and shape
# information
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes

# Converts Conv to Im2Col and MatMul with extracted standalone bias op
from qonnx.transformation.lower_convs_to_matmul import LowerConvsToMatMul

# Transposes the initializer tensors of a Quant node instead of having a
# standalone Transpose following
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit

# Range information structure for seeding the range analysis for converting
# quantized activations to MultiThreshold
from qonnx.util.range_analysis import RangeInfo

# FINN dataflow builder configuration
from finn.builder.build_dataflow_config import DataflowBuildConfig, VerificationStepType

# FINN verification after build/graph transformation steps
from finn.builder.build_dataflow_steps import verify_step

# Detects the attention pattern and converts to hardware custom op
from finn.transformation.fpgadataflow.attention import (
    AbsorbMultiThresholdIntoScaledDotProductAttention,
    InferScaledDotProductAttention,
)

# Mult-Head Attention support
from finn.transformation.fpgadataflow.attention_heads import (
    InferMultiHeads,
    MoveMergeMultiHeadsPastMultiThreshold,
    MoveSplitMultiHeadsPastMultiThreshold,
    UnrollMultiHeadAttention,
)
from finn.transformation.fpgadataflow.compile_cppsim import CompileCppSim

# Converts (infers) ONNX and QONNX nodes to FINN hardware CustomOps
from finn.transformation.fpgadataflow.convert_to_hw_layers import (
    InferConcatLayer,
    InferElementwiseBinaryOperation,
    InferLookupLayer,
    InferSplitLayer,
    InferSqueeze,
    InferUnsqueeze,
    InferVectorVectorActivation,
)
from finn.transformation.fpgadataflow.hlssynth_ip import HLSSynthIP

# Inserts data-width converter and FIFO nodes into the model graph
from finn.transformation.fpgadataflow.insert_dwc import InsertDWC
from finn.transformation.fpgadataflow.insert_fifo import InsertFIFO

# Transformations preparing the operators for synthesis and simulation
from finn.transformation.fpgadataflow.prepare_cppsim import PrepareCppSim
from finn.transformation.fpgadataflow.prepare_ip import PrepareIP

# Converts fork-nodes to ReplicateStream hardware operator
from finn.transformation.fpgadataflow.replicate_stream import InferReplicateStream
from finn.transformation.fpgadataflow.set_exec_mode import SetExecMode

# Splitting and removing of FIFOs from the model graph
from finn.transformation.fpgadataflow.set_fifo_depths import RemoveShallowFIFOs, SplitLargeFIFOs

# Graph transformation setting the folding, i.e., parallelization configuration
from finn.transformation.fpgadataflow.set_folding import SetFolding

# Specializes each layer's implementation style: HLS or RTL implementation
from finn.transformation.fpgadataflow.specialize_layers import SpecializeLayers

# Standard QONNX to FINN conversion function
from finn.transformation.qonnx.convert_qonnx_to_finn import ConvertQONNXtoFINN

# Folds quantizers into weight tensor initializers, needed for lowering
# convolutions to MatMuls
from finn.transformation.qonnx.fold_quant_weights import FoldQuantWeights
from finn.transformation.qonnx.quant_act_to_multithreshold import default_filter_function_generator

# Cleanup transformation getting rid of 3d data layout
from finn.transformation.squeeze import Squeeze
from finn.transformation.streamline.absorb import (
    AbsorbAddIntoMultiThreshold,
    AbsorbSignBiasIntoMultiThreshold,
)

# FINN streamlining transformations fusing/collapsing operations of the same
# kind
from finn.transformation.streamline.collapse_repeated import CollapseRepeatedTranspose

# FINN streamlining transformations removing nodes without real effect from the
# graph
from finn.transformation.streamline.remove import RemoveIdentityReshape, RemoveIdentityTranspose

# FINN streamlining transformations reordering the graph
from finn.transformation.streamline.reorder import (
    MoveMulPastAdd,
    MoveSqueezePastMatMul,
    MoveSqueezePastMultiThreshold,
    MoveTransposePastEltwise,
    MoveTransposePastFork,
    MoveTransposePastJoinAdd,
    MoveTransposePastJoinConcat,
    MoveTransposePastJoinMul,
    MoveTransposePastSplit,
)
from finn.transformation.streamline.streamline_plus import StreamlinePlus as Streamline


# Prepares the graph to be consumed by FINN:
# 1. Some graph cleanup removing unused tensors, nodes without effect and
#  folding constants, i.e., collapsing chains of operations on constant tensors
# 2. Lowers some "more complex" operations: converts Conv and Gemm to MatMul and
#  BatchNorm to Mul and Add operations followed by some necessary cleanup
# 3. Converts all QONNX Quant nodes to MultiThreshold operations which can
#  absorb scales and biases during streamlining
def prepare_graph(range_info: RangeInfo):
    """
    Prepares the graph to be consumed by FINN.

    Performs graph cleanup, lowers complex operations to basic ones, and converts
    QONNX Quant nodes to MultiThreshold operations.

    Args:
        range_info: Range information for seeding range analysis.

    Returns:
        A wrapped transformation step function that performs graph preparation.
    """

    # Wrap the actual transformation/build step function
    def step_prepare_graph(model: ModelWrapper, cfg: DataflowBuildConfig):
        """
        The actual transformation/build step function that prepares the graph.

        Args:
            model: The model wrapper containing the ONNX graph.
            cfg: The dataflow build configuration.

        Returns:
            The transformed model with prepared graph.
        """
        # Exhaustively apply the set of cleanup transformations
        model = model.transform(
            ComposedTransformation(
                [
                    # Adds shape and datatype annotations to all tensors in this graph
                    InferDataTypes(),
                    InferShapes(),
                    # Cleanup the graph by removing redundant, unnecessary and constant
                    # nodes and tensors and give unique names to everything remaining
                    GiveUniqueNodeNames(),
                    GiveReadableTensorNames(),
                    RemoveStaticGraphInputs(),
                    RemoveUnusedTensors(),
                    GiveUniqueParameterTensors(),
                    FoldConstants(),
                    # Remove unnecessary shape and layout transformations
                    RemoveIdentityReshape(),
                    RemoveIdentityTranspose(),
                    # Redo shape and datatype annotations after removing nodes and
                    # tensors
                    InferShapes(),
                    InferDataTypes(),
                ]
            )
        )
        # If configured, run a verification of the transformed model on some
        # sample inputs
        if VerificationStepType.TIDY_UP_PYTHON in cfg._resolve_verification_steps():  # noqa
            verify_step(model, cfg, "tidied_up_python", need_parent=False)
        # Exhaustively apply the lowering transformations
        model = model.transform(
            ComposedTransformation(
                [
                    # Moves the bias input to the Conv operator as a separate Add node
                    # behind the Conv node
                    ExtractBiasFromConv(),
                    # Converts Gemm nodes to MatMul (+ bias)
                    GemmToMatMul(),
                    # Need to do some constant and weight folding first
                    FoldConstants(),
                    FoldTransposeIntoQuantInit(),
                    FoldQuantWeights(),
                    # Annotate the graph with shape and data type information
                    InferShapes(),
                    InferDataTypes(),
                    # Converts Conv layers to MatMul
                    LowerConvsToMatMul(),
                    # Converts BatchNorm to affine scale and bias
                    BatchNormToAffine(),
                    # Annotate the graph with shape and data type information
                    InferShapes(),
                    InferDataTypes(),
                ]
            )
        )
        # If configured, run a verification of the transformed model on some
        # sample inputs
        if VerificationStepType.QONNX_TO_FINN_PYTHON in cfg._resolve_verification_steps():  # noqa
            verify_step(model, cfg, "lowered_python", need_parent=False)

        # Apply the standard QONNX to FINN conversion step to convert the
        # remaining quantizers not yet covered by the new range analysis based
        # method
        model = model.transform(
            ConvertQONNXtoFINN(
                filter_function=default_filter_function_generator(
                    max_multithreshold_bit_width=cfg.max_multithreshold_bit_width
                )
            )
        )
        # If configured, run a verification of the transformed model on some
        # sample inputs
        if VerificationStepType.QONNX_TO_FINN_PYTHON in cfg._resolve_verification_steps():  # noqa
            verify_step(model, cfg, "prepared_graph_python", need_parent=False)
        # Return the transformed model
        return model

    # Return the wrapped transformation step function
    return step_prepare_graph


# Applies the custom set of exhaustive streamlining transformations, also taking
# special topology like attention, residuals, splits and transposes into account
def step_streamline(model: ModelWrapper, cfg: DataflowBuildConfig):
    """
    Applies custom exhaustive streamlining transformations.

    Takes special topology like attention, residuals, splits and transposes into account.

    Args:
        model: The model wrapper containing the ONNX graph.
        cfg: The dataflow build configuration.

    Returns:
        The streamlined model.
    """
    # These should not be applied exhaustively with the other streamlining
    # transformations to not end up in cycles.
    # Note: This is essential to allow some Add operations to be
    # absorbed by the next round's AbsorbSignBiasIntoMultiThreshold
    model = model.transform(MoveMulPastAdd())
    model = model.transform(AbsorbSignBiasIntoMultiThreshold())
    # Exhaustively apply the following set of transformations to streamline the
    # graph with the overall goal of collecting scales and biases in front of
    # MultiThreshold operations or, alternatively, at the end of the graph.
    # Note: Contains some sets of nested exhaustive transformations meant for
    # particular architectural patterns, e.g., residual topologies.
    model = model.transform(Streamline())
    # If configured, run a verification of the transformed model on some
    # sample inputs
    if VerificationStepType.STREAMLINED_PYTHON in cfg._resolve_verification_steps():  # noqa
        verify_step(model, cfg, "streamlined_python", need_parent=False)
    # Return the transformed model
    return model


# Converts scaled dot-product attention operations to FINN hardware operations
# Note: This includes some necessary cleanup after converting the pattern, in
# particular squeezing the data layouts throughout the graph
def step_convert_attention_to_hw(model: ModelWrapper, _: DataflowBuildConfig):
    """
    Converts scaled dot-product attention operations to FINN hardware operations.

    Includes necessary cleanup after converting the pattern, in particular squeezing
    the data layouts throughout the graph.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with attention and multi-heads mapped to hardware operators.
    """
    # Try to infer reshaping of attention heads
    model = model.transform(InferMultiHeads())  # noqa: Duplicate
    # Try to mode the mult-head splitting past the multi thresholds
    model = model.transform(MoveSplitMultiHeadsPastMultiThreshold())
    # Moving multi-head splitting past multi thresholds might enable absorbing
    # adds into thresholds once again
    model = model.transform(AbsorbAddIntoMultiThreshold())
    # Try to infer a ScaledDotProductAttention custom op
    model = model.transform(InferScaledDotProductAttention())
    # Parallelize attention head in the onnx graph
    model = model.transform(UnrollMultiHeadAttention())
    # Swap the order of merging the multi heads and applying thresholds
    model = model.transform(MoveMergeMultiHeadsPastMultiThreshold())
    # If applicable, absorb the final thresholds into the attention operator
    model = model.transform(AbsorbMultiThresholdIntoScaledDotProductAttention())
    # Squeeze (i.e., remove dimensions of size 1) the data layouts throughout
    # the graph to treat the time dimension as the batch dimension for all MVU
    # and Threshold operators
    model = model.transform(Squeeze())
    # Squeezing might have turned further transpose and reshape operations into
    # identities (those which just swapped around the dimensions of size 1)
    model = model.transform(
        ComposedTransformation(
            [
                # Move transposes around to some place where they could be removed
                # later, i.e., where they collapse into identities
                MoveTransposePastFork(),
                MoveTransposePastSplit(),
                MoveTransposePastJoinConcat(),
                MoveTransposePastEltwise(),
                MoveTransposePastJoinMul(),
                MoveTransposePastJoinAdd(),
                CollapseRepeatedTranspose(),
                # Remove identity shape/layout transformations
                RemoveIdentityTranspose(),
                RemoveIdentityReshape(),
                # Squeeze operators can be moved past MatMuls and thresholding
                MoveSqueezePastMatMul(),
                MoveSqueezePastMultiThreshold(),
            ]
        )
    )
    # Squeezing might enable absorbing adds into thresholds once again
    model = model.transform(AbsorbAddIntoMultiThreshold())
    # If applicable, absorb the final thresholds into the attention operator
    #   Note: Might be applicable again after squeezing a transpose away
    model = model.transform(AbsorbMultiThresholdIntoScaledDotProductAttention())
    # We should do another round of streamlining to be sure and support more
    # general architectural patterns, we are not aware of yet...
    model = model.transform(Streamline())
    # Convert Squeeze and Unsqueeze operators to hardware operations
    model = model.transform(InferSqueeze())
    model = model.transform(InferUnsqueeze())
    # Return the model with attention and multi-heads mapped to hardware
    # operators
    return model


# Function running the transformations to convert elementwise binary operations
# to their hardware implementations
def step_convert_elementwise_binary_to_hw(model: ModelWrapper, _):
    """
    Converts elementwise binary operations to their hardware implementations.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with elementwise operations converted to hardware operators.
    """
    # Convert elementwise operations to hardware operators
    #   Note: Do not convert the final Mul operator at the output
    return model.transform(
        InferElementwiseBinaryOperation(InferElementwiseBinaryOperation.reject_output_dequant)
    )


# Converts Split and Concat operations to hardware custom operators
def step_convert_split_concat_to_hw(model: ModelWrapper, _):
    """
    Converts Split and Concat operations to hardware custom operators.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with Split and Concat operations converted to hardware layers.
    """
    return model.transform(InferSplitLayer()).transform(InferConcatLayer())


# Function running the transformations to convert Gather, i.e., index lookup,
# nodes to their hardware implementations
def step_convert_lookup_to_hw(model: ModelWrapper, _):
    """
    Converts Gather (index lookup) nodes to their hardware implementations.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with Gather operations converted to Lookup layers.
    """
    # Iterate all nodes in the graph keeping track of the index
    for index, node in enumerate(model.graph.node):
        # If this is a Gather node, force the input (index) type annotation
        if node.op_type == "Gather":
            # Force input type to unsigned 64-bit integer for now
            model.set_tensor_datatype(node.input[1], DataType["UINT64"])
            # Workaround:
            # Also force global input to UINT64 to avoid type inference to reset it to FLOAT32
            # TODO: Address properly by either
            # 1) Annotating the global input with the correct datatype (at export)
            # 2) Make ModelWrapper.get_tensor_datatype() fall back to container type, not of FP32
            model.set_tensor_datatype(model.graph.input[0].name, DataType["UINT64"])
            # Get the value info for the input tensor to have access to the ONNX
            # datatype of the tensor
            value_info = model.get_tensor_valueinfo(node.input[1])
            # Force the container datatype of the input to be a float
            value_info.type.tensor_type.elem_type = 1
    # Convert Gather to Lookup layers
    return model.transform(InferLookupLayer())


# Converts depth-wise convolution to hardware operator calling the
# InferVectorVectorActivation transformation
def step_convert_depth_wise_to_hw(model: ModelWrapper, _: DataflowBuildConfig):
    """
    Converts depth-wise convolution to hardware operator.

    Calls the InferVectorVectorActivation transformation.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with depth-wise convolutions converted to hardware operators.
    """
    return model.transform(InferVectorVectorActivation())


# Function running the InferReplicateStream transformation
def step_replicate_streams(model: ModelWrapper, _):
    """
    Runs the InferReplicateStream transformation.

    Properly replicates the stream feeding the query, key and value projections.

    Args:
        model: The model wrapper containing the ONNX graph.
        _: The dataflow build configuration (unused).

    Returns:
        The model with replicated streams for attention projections.
    """
    # Properly replicate the stream feeding the query, key and value projections
    return model.transform(InferReplicateStream())


# Custom step for setting the parallelism to meet the target of T^2 cycles per
# sequence
def set_target_parallelization(seq_len: int, emb_dim: int):  # noqa: emb_dim
    """
    Custom step for setting the parallelism to meet the target of T^2 cycles per sequence.

    Args:
        seq_len: The sequence length.
        emb_dim: The embedding dimension (unused but kept for compatibility).

    Returns:
        A wrapped transformation step function that configures parallelization.
    """

    # The wrapping function is a generator and this is the actual build step
    # function taking the model and build configuration
    def step_set_target_parallelization(model: ModelWrapper, cfg: DataflowBuildConfig):
        """
        The actual build step function that sets target parallelization.

        Args:
            model: The model wrapper containing the ONNX graph.
            cfg: The dataflow build configuration.

        Returns:
            The model with configured parallelization to meet T^2 cycles target.
        """
        # Run over all nodes in the model graph to look for attention operators,
        # which are currently not handled by the SetFolding transformation
        for index, node in enumerate(model.graph.node):
            # Only handle attention operations here
            if node.op_type == "ScaledDotProductAttention_hls":
                # Convert this to the custom-op instance for easy access to node
                # attributes
                inst = getCustomOp(node)
                # Set the sequence and embedding dimension folding to meet the
                # T^2 cycles target, i.e., fully parallel along the embedding
                # dimension and fully sequential along the sequence dimension
                inst.set_nodeattr("EmbFold", 1)
                inst.set_nodeattr("SeqFold", seq_len)
        # Apply the built-in folding configuration transformation with the
        # T^2 target cycles
        model = model.transform(
            SetFolding(seq_len**2, cfg.mvau_wwidth_max, cfg.folding_two_pass_relaxation)
        )
        # TODO: Extract the folding configuration
        # Return the model with configured parallelization
        return model

    # Return the wrapped build step function
    return step_set_target_parallelization


# Applies configuration dictionary to the model graph
class ApplyConfig(Transformation):
    """
    Applies configuration dictionary to the model graph.

    Transformation that applies operator-specific and node-specific configurations
    from a dictionary to the model graph nodes.
    """

    # Initializes the transformation with the configuration dictionary

    def __init__(self, config):
        """
        Initializes the transformation with the configuration dictionary.

        Args:
            config: Configuration dictionary containing default and node-specific settings.
        """
        # Initialize the transformation base class
        super().__init__()
        # Register the configuration dictionary to be used in apply()
        self.config = config

    # Applies the transform to a whole model graph
    def apply(self, model: ModelWrapper):  # noqa
        """
        Applies the transform to a whole model graph.

        Args:
            model: The model wrapper containing the ONNX graph.

        Returns:
            Tuple of (transformed_model, graph_modified_flag).
        """
        # Get the model graph out of the model wrapper object
        graph = model.graph
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(graph.node):
            # A node should not be named "defaults"...
            assert node.name != "defaults", "Node has reserved name 'defaults'"
            # Convert this to the custom-op instance for easy access to node
            # attributes
            inst = getCustomOp(node)
            # Apply the per operator type default configurations to the node
            if node.op_type in self.config["defaults"]:
                # Run over all default options to be applied to this node
                for key, value in self.config["defaults"][node.op_type].items():
                    # Set the nodes attribute to the default option value
                    inst.set_nodeattr(key, value)
            # If there is an individual, node-specific configuration apply
            # this next, potentially overriding the defaults set above
            if node.name in self.config:
                # Run over all node-specific options to be applied to this
                # node
                for key, value in self.config[node.name].items():
                    # Set the nodes attribute to the option value
                    inst.set_nodeattr(key, value)
        # Return model with configuration applied
        # Note: Do not consider this as modifying the graph. This does not have
        # to be reapplied multiple times.
        return model, False


# Custom build step trying to set appropriate FIFO sizes for the transformer
def set_fifo_depths(seq_len: int, emb_dim: int, uram_threshold: int = 32):  # noqa: emb_dim
    """
    Custom build step to set appropriate FIFO sizes for the transformer.

    Args:
        seq_len: The sequence length.
        emb_dim: The embedding dimension (unused but kept for compatibility).
        uram_threshold: Threshold depth for using URAM instead of BRAM (default: 32).

    Returns:
        A wrapped transformation step function that configures FIFO depths.
    """

    # The wrapping function is a generator and this is the actual build step
    # function taking the model and build configuration
    def step_set_fifo_depths(model: ModelWrapper, cfg: DataflowBuildConfig):
        """
        The actual build step function that sets FIFO depths.

        Configures appropriate FIFO sizes for transformer operations, with special
        handling for attention and residual branches.

        Args:
            model: The model wrapper containing the ONNX graph.
            cfg: The dataflow build configuration.

        Returns:
            The model with configured FIFO depths and hardware attributes.
        """
        # Run over all nodes in the model graph
        for index, node in enumerate(model.graph.node):
            # Convert this to the custom-op instance for easy access to node
            # attributes
            inst = getCustomOp(node)
            # Extract the FIFO depths configuration of the node
            in_depths = inst.get_nodeattr("inFIFODepths")
            out_depths = inst.get_nodeattr("outFIFODepths")

            # Number of inputs and outputs to/from the node
            num_inputs = len(node.input)
            num_outputs = len(node.output)

            # If the input/output has only default configurations, fill with as
            # many shallow FIFOs as there are inputs, to avoid later problems
            # with to few FIFO depths specified
            if in_depths == [2] and num_inputs > 1:
                in_depths = num_inputs * [2]
            if out_depths == [2] and num_outputs > 1:
                out_depths = num_outputs * [2]

            # Special case: Attention needs properly sized input FIFOs
            if node.op_type == "ScaledDotProductAttention_hls":
                # Each folded input stream needs to be buffered completely
                # TODO: Not exactly sure whether this is always correct or just
                #  the worst-case
                in_depths = [inst.get_number_input_values(i) for i in range(num_inputs)]
                # Note: No special treatment of the output FIFO
                # out_depths = ...

            # Special case: Adding residual branches needs to buffer the inputs
            # to avoid deadlocks if one branch is running faster/slower
            if node.op_type == "ElementwiseAdd_hls":
                # Only relevant if for join-node operations, i.e., node actually
                # consumes two branches, potentially operating at a different
                # rate
                if model.is_join_node(node):
                    # Set both inputs to buffer as many cycles as we target for
                    # the attention operations, i.e., the T^2 cycles per
                    # sequence target
                    # TODO: Not exactly sure whether this is always correct or
                    #  just the worst-case
                    # TODO: Currently we do not really have a reliable way of
                    #  figuring out which of the two is the longer/deeper branch
                    #  in terms of cycles to set a corresponding buffer only to
                    #  the shorter branch.
                    in_depths = [seq_len**2, seq_len**2]
                    # Note: No special treatment of the output FIFO
                    # out_depths = ...

            # Set the updated FIFO depths attributes
            inst.set_nodeattr("inFIFODepths", in_depths)
            inst.set_nodeattr("outFIFODepths", out_depths)

        # The following partially mirrors (or even copies from) the build-in
        # step_set_fifo_depths using only manual FIFO depths and our YAML-based
        # folding configuration.

        # Insert data-width converters
        model = model.transform(InsertDWC())
        # Insert FIFOs between all operators (inserts shallow, depths 2 FIFOs if
        # no other depth is specified)
        model = model.transform(InsertFIFO(create_shallow_fifos=True))
        # Specialize the implementation variant of the (newly added FIFO) layers
        model = model.transform(SpecializeLayers(cfg._resolve_fpga_part()))  # noqa: Access _ method
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())

        # Only applies if a configuration file is given
        if cfg.folding_config_file is not None:
            # Load the configuration dictionary form YAML file
            with open(cfg.folding_config_file, "r") as file:
                # Load YAML string
                config = yaml.safe_load(file)
                # Assign unique names to the nodes which can be matched by
                # individual per-node configuration options
                model = model.transform(GiveUniqueNodeNames())
                # Apply the configuration dictionary to the model graph
                model = model.transform(ApplyConfig(config))

        # Run over all nodes in the model graph once again to modify the
        # inserted FIFOs
        # Note: This overwrites the folding configuration...
        # TODO: Find a better way to handle this
        for index, node in enumerate(model.graph.node):
            # Modify all RTL FIFO operators
            if node.op_type == "StreamingFIFO_rtl":
                # Convert this to the custom-op instance for easy access to node
                # attributes
                inst = getCustomOp(node)
                # Check the depth of the FIFO: If this is not a shallow FIFO,
                # implement this via the vivado strategy in URAM
                if inst.get_nodeattr("depth") >= uram_threshold:
                    # Change the implementation style to vivado
                    inst.set_nodeattr("impl_style", "vivado")
                    # Set the resource type for the memory to URAM
                    inst.set_nodeattr("ram_style", "ultra")

        # Hardware attributes to be extracted from each node
        hw_attrs = {
            "PE",
            "SIMD",
            "parallel_window",
            "ram_style",
            "ram_style_thresholds",
            "ram_style_mask",
            "depth",
            "impl_style",
            "resType",
            "mac_resource",
            "mem_mode",
            "runtime_writeable_weights",
            "inFIFODepths",
            "outFIFODepths",
            "depth_trigger_uram",
            "depth_trigger_bram",
        }

        # Start collecting the configuration from the model graph as a
        # dictionary
        config = {"defaults": {}}
        # Iterate all nodes in the graph keeping track of the index
        for index, node in enumerate(model.graph.node):
            # Convert this to the custom-op instance for easy access to node
            # attributes
            inst = getCustomOp(node)
            # Prepare the node-specific configuration entry for this node
            config[node.name] = {}
            # Collect attribute values for all specified hardware attributes
            for key in hw_attrs:
                # Some hardware attributes may not be present for all nodes or
                # op-types, this will be signaled via exception
                try:
                    # Try extracting the configuration value from the node
                    # custom-op instance
                    config[node.name][key] = inst.get_nodeattr(key)
                # Missing attributes are signaled va AttributeError
                except AttributeError:
                    # Can be safely ignored here
                    pass
            # Cleanup: If no attribute is present for this node, there is no
            # need to keep this in the configuration dictionary as there is
            # nothing to be restored later
            if not config[node.name]:
                # Remove the entry form the configuration dictionary
                del config[node.name]

        # Create/Open a YAML file to store the configuration for later reuse
        # TODO: make consistent with .json report in default step
        with open(cfg.output_dir + "/report/final_hw_config.yaml", "w") as file:
            # Store the configuration dictionary as YAML code
            yaml.safe_dump(config, file)

        # Perform FIFO splitting and shallow FIFO removal only after the final
        # config file has been written. Otherwise, since these transforms may
        # add/remove FIFOs, we get name mismatch problems when trying to reuse
        # the final config.
        if cfg.split_large_fifos:
            model = model.transform(SplitLargeFIFOs())
        model = model.transform(RemoveShallowFIFOs())

        # generate a dedicated report about final FIFO sizes
        fifo_info = {}
        fifo_info["fifo_depths"] = {}
        fifo_info["fifo_sizes"] = {}
        total_fifo_size = 0
        for node in model.get_nodes_by_op_type("StreamingFIFO_rtl"):
            node_inst = getCustomOp(node)
            fifo_info["fifo_depths"][node.name] = node_inst.get_nodeattr("depth")
            fifo_info["fifo_sizes"][
                node.name
            ] = node_inst.get_instream_width() * node_inst.get_nodeattr("depth")
            total_fifo_size += fifo_info["fifo_sizes"][node.name]
        fifo_info["total_fifo_size_kB"] = int(total_fifo_size / 8.0 / 1000.0)

        with open(cfg.output_dir + "/report/fifo_sizing.json", "w") as f:
            json.dump(fifo_info, f, indent=2)

        # After FIFOs are ready to go, call PrepareIP and HLSSynthIP again
        # this will only run for the new nodes (e.g. FIFOs and DWCs)
        model = model.transform(
            PrepareIP(cfg._resolve_fpga_part(), cfg._resolve_hls_clk_period())  # noqa
        )
        model = model.transform(HLSSynthIP())

        # Return the model with configured parallelization
        return model

    # Return the wrapped build step function
    return step_set_fifo_depths


# Custom step applying our custom format of folding configuration to the graph
def step_apply_folding_config(model: ModelWrapper, cfg: DataflowBuildConfig):
    """
    Custom step applying custom format of folding configuration to the graph.

    Loads and applies folding configuration from YAML file if specified in the
    build configuration.

    Args:
        model: The model wrapper containing the ONNX graph.
        cfg: The dataflow build configuration.

    Returns:
        The model with folding configuration applied.
    """
    # Only applies if a configuration file is given
    if cfg.folding_config_file is not None:
        # Load the configuration dictionary form YAML file
        with open(cfg.folding_config_file, "r") as file:
            # Load YAML string
            config = yaml.safe_load(file)
            # Assign unique names to the nodes which can be matched by
            # individual per-node configuration options
            model = model.transform(GiveUniqueNodeNames())
            # Apply the configuration dictionary to the model graph
            model = model.transform(ApplyConfig(config))
    # If configured, run a verification of the transformed model on some sample
    # inputs
    if VerificationStepType.FOLDED_HLS_CPPSIM in cfg._resolve_verification_steps():  # noqa
        # Prepare C++ Simulation for verification
        model = model.transform(PrepareCppSim())
        model = model.transform(CompileCppSim())
        model = model.transform(SetExecMode("cppsim"))
        # Execute a verification step of the model with inputs specified in
        # build configuration
        verify_step(model, cfg, "folded_hls_cppsim", need_parent=True)

    # Return model with configuration applied
    return model



from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.change_3d_tensors_to_4d import Change3DTo4DTensors
from qonnx.transformation.general import GiveUniqueNodeNames

import finn.transformation.fpgadataflow.convert_to_hw_layers as to_hw
import finn.transformation.streamline.absorb as absorb
from finn.builder.build_dataflow_config import DataflowBuildConfig


def step_pre_streamline(model: ModelWrapper, cfg: DataflowBuildConfig):
    model = model.transform(Change3DTo4DTensors())
    model = model.transform(absorb.AbsorbScalarMulAddIntoTopK())
    return model


def step_convert_final_layers(model: ModelWrapper, cfg: DataflowBuildConfig):
    model = model.transform(to_hw.InferChannelwiseLinearLayer())
    model = model.transform(to_hw.InferLabelSelectLayer())
    model = model.transform(GiveUniqueNodeNames())
    return model
